/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef LEVEL_GODUNOV_H
#define LEVEL_GODUNOV_H

// original design        : Phil Colella
// original implementation: Brian Van Straalen
// implementation date    : Mon Apr 12, 1999

#include "FArrayBox.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "LevequeCorrector.H"
#include "PiecewiseLinearFillPatch.H"
#include "ExtrapFillPatch.H"
#include "RadiusMask.H"
#include "LevelFluxRegister.H"
#include "PhysIBC.H"

/// Level Godunov
/**
 */
class LevelGodunov
{
public:
  /// Default constructor
  /**
     Object requires define(..) to be called before all other functions.
  */
  LevelGodunov();

  /// Destructor
  /**
     destroys all objects created by define(..). Passed in
     data references of define(..) are left alone.
  */
  ~LevelGodunov();

  /// Actual constructor.
  /**
     For the coarsest level, an empty DisjointBoxLayout
     is passed in for coaserDisjointBoxLayout.
     Inside the routine, we cast away const-ness on the data members
     for the assignment.  The arguments passed in are maintained const.
     (coding standards).  nRefineCrse is the refinement ratio between
     this level and the next coarser level.
  */
  void define(const DisjointBoxLayout& a_thisDisjointBoxLayout,
              const DisjointBoxLayout& a_coarserDisjointBoxLayout,
              const Box&               a_DProblem,
              const int&               a_numGhost,
              const int&               a_nRefine,
              const Real&              a_dx,
              const RefCountedPtr<PhysIBC>&   a_bc);


  /// Actual constructor.
  /**
     For the coarsest level, an empty DisjointBoxLayout
     is passed in for coaserDisjointBoxLayout.
     Inside the routine, we cast away const-ness on the data members
     for the assignment.  The arguments passed in are maintained const.
     (coding standards).  nRefineCrse is the refinement ratio between
     this level and the next coarser level.
  */
  void define(const DisjointBoxLayout& a_thisDisjointBoxLayout,
              const DisjointBoxLayout& a_coarserDisjointBoxLayout,
              const ProblemDomain&     a_DProblem,
              const int&               a_numGhost,
              const int&               a_nRefine,
              const Real&              a_dx,
              const RefCountedPtr<PhysIBC>&   a_bc);

  /// Take one timestep for this grid level.
  /**
     For the coarsest level, UcoarseOld and UCoarseNew are empty
     LevelData<FArrayBox> objects.  Also, either UCoarseOld or
     UCoarseNew might be empty to indicate that t(nf) = t(nc) the
     one grid is at the current time and interpolation is not required
     for boundary condition generation.  U must be defined on the same
     DisjointBoxLayouts as were used in define(...)
  */
  Real step(LevelData<FArrayBox>&       a_U,
            LevelFluxRegister&          a_myFluxRegister,
            LevelFluxRegister&          a_fineFluxRegister,
            const LevelData<FArrayBox>& a_UCoarseOld,
            const LevelData<FArrayBox>& a_UCoarseNew,
            const Real&                 a_time,
            const Real&                 a_TCold,
            const Real&                 a_TCNew,
            const Real&                 a_dt
            );

  /// Get maximum wave speed
  Real getMaxWaveSpeed(const LevelData<FArrayBox>& a_state);

protected:
  /// Get flattening
  void getFlattening(const FArrayBox& a_q,
                     const DataIndex  a_di,
                     const int        a_dir,
                     FArrayBox&       a_flatCoef) const;

  ///
  /**
     Computes the primitive variables at each grid point from the
     conservative variables stored in state, stores them in q.
  */
  void consToPrim(const FArrayBox& a_conState,
                  FArrayBox&       a_primState) const;

  ///
  /**
     Compute the divergence of Unormal from primitive variables in q.
     divU is defined on the edges between the cell centers in the sweep
     direction.
  */
  void divU(const FArrayBox& a_q,
            const Box&       a_box,
            const int        a_dir,
            FArrayBox&       a_divU) const;

  ///
  /**
     Compute the slope of q in the dir direction and store the
     result in dq.  q and dq are cell-centered.
  */
  void slope(const FArrayBox& a_q,
             const DataIndex  a_di,
             const int        a_dir,
             FArrayBox&       a_dq) const;

  ///
  /**
     Compute edge-extrapolated values of q,  store into qlo, qhi
     qlo and qhi are defined inside of normalPred, they come
     out of this function as edge-shifted boxes.  bvs
  */
  void normalPred(const FArrayBox& a_q,
                  const FArrayBox& a_dq,
                  const Real&      a_dtime,
                  const int        a_dir,
                  FArrayBox&       a_qlo,
                  FArrayBox&       a_qhi) const;

  ///
  /**
     Take edge-centered low and high values as input for the Riemann
     problem at each edge.  Store result in qGdnv
  */
  void riemann(const FArrayBox& a_qleft,
               const FArrayBox& a_qright,
               const int        a_dir,
               FArrayBox&       a_qGdnv) const;

  /// Compute fluxes
  void fluxCalc(const FArrayBox& a_qGdnv,
                const int        a_dir,
                FArrayBox&       a_flux) const;

  /// Apply artificial viscosity to fluxes
  void artificialViscosity(const FArrayBox& a_qGdnv,
                           const FArrayBox& a_conState,
                           const FArrayBox& a_divU,
                           const int        a_dir,
                           FArrayBox&       a_flux) const;

  /// Update state
  void updateState(const FArrayBox& a_flux,
                   const int        a_dir,
                   const Real       a_dtime,
                   const Box&       a_box,
                   FArrayBox&       a_conState) const;


  bool                     m_defined;
  const DisjointBoxLayout  m_ba;
  RefCountedPtr<PhysIBC>   m_bc;
  LevequeCorrector         m_leveque;
  int                      m_numGhost;

  PiecewiseLinearFillPatch m_patcher;
  ExtrapFillPatch          m_extrap_filler;
  Real                     m_dx;
  ProblemDomain            m_DProblem;
  Box                      m_DProblemNodal;
  int                      m_refine_coarse;
  int                      m_numCons;
  int                      m_numPrim;
  int                      m_numSlopes;
  int                      m_numFluxes;

  // see RadiusMask.H for this class description
  RadiusMask               m_mask;

  ///
  /**
     Data member used in some asserts for box type-checking
     (since the compiler can't type check box types for us.
  */
  IndexType                m_edgeType[SpaceDim];
};

#endif
