/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iomanip>

#include "parstream.H"

#include "Box.H"
#include "Vector.H"
#include "IntVectSet.H"
#include "FArrayBox.H"
#include "DisjointBoxLayout.H"
#include "LayoutIterator.H"
#include "HDF5.H"
#include "SPMD.H"
#include "LoadBalance.H"
#include "LevelFluxRegister.H"
#include "ProblemDomain.H"
#include "BoxIterator.H"
#include "HOExtrapBC.H"
#include "Extrap0thOrderBC.H"
#include "AMRSolver.H"
#include "AMRIO.H"

#include "AMRLevel.H"
#include "AMRAdvectDiffuse.H"

#include "computeSum.H"

#include "PatchGodunovF_F.H"
#include "AdvectTestF_F.H"


// tolerance for checking equality between real numbers
#ifdef CH_USE_DOUBLE
#define TIME_EPS 1.0e-10
#else
#define TIME_EPS 1.0e-5
#endif


// Constructor
AMRAdvectDiffuse::AMRAdvectDiffuse()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse default constructor" << endl;
  }

  m_cfl = 0.8;
  m_domainLength = 1.0;
  m_refineThresh = 0.2;
  m_initial_dt_multiplier = 0.1;
  m_patchGodunovFactory = NULL;
  m_patchGodunov = NULL;
  m_diffusionSolverPtr = NULL;
  m_doImplicitReflux = false;
}

// Destructor
AMRAdvectDiffuse::~AMRAdvectDiffuse()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse destructor" << endl;
  }

  // Get rid of the patch integrators (factory and object)
  if (m_patchGodunovFactory != NULL)
  {
    delete m_patchGodunovFactory;
  }

  if (m_patchGodunov != NULL)
  {
    delete m_patchGodunov;
  }

  if (m_diffusionSolverPtr != NULL)
    {
      delete m_diffusionSolverPtr;
    }
}

// This instance should never get called - historical
void AMRAdvectDiffuse::define(AMRLevel*  a_coarserLevelPtr,
                                   const Box& a_problemDomain,
                                   int        a_level,
                                   int        a_refRatio)
{

  MayDay::Error("AMRAdvectDiffuse::define -\n\tShould never be called with a Box for a problem domain");
}

// Define new AMR level
void AMRAdvectDiffuse::define(AMRLevel*            a_coarserLevelPtr,
                              const ProblemDomain& a_problemDomain,
                              int                  a_level,
                              int                  a_refRatio)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::define " << a_level << endl;
  }

  // Call inherited define
  AMRLevel::define(a_coarserLevelPtr,
                   a_problemDomain,
                   a_level,
                   a_refRatio);
  
  
  // Get setup information from the next coarser level
  int nRefCrse = -1;
  const DisjointBoxLayout* crseGridsPtr = NULL;

  if (a_coarserLevelPtr != NULL)
    {
      AMRAdvectDiffuse* amrGodPtr = dynamic_cast<AMRAdvectDiffuse*>(a_coarserLevelPtr);
      
      if (amrGodPtr != NULL)
        {
          m_cfl = amrGodPtr->m_cfl;
          m_domainLength = amrGodPtr->m_domainLength;
          m_refineThresh = amrGodPtr->m_refineThresh;
          m_tagBufferSize = amrGodPtr->m_tagBufferSize;
          m_backgroundVelocity = amrGodPtr->m_backgroundVelocity;
          m_nu = amrGodPtr->m_nu;
          nRefCrse = amrGodPtr->m_ref_ratio;
          crseGridsPtr = &(amrGodPtr->m_grids);
          m_diffusionSolverPtr = amrGodPtr->m_diffusionSolverPtr->new_heatSolver();
          m_doImplicitReflux = amrGodPtr->m_doImplicitReflux;
          m_diffusion_solver_tol = amrGodPtr->m_diffusion_solver_tol;
        }
      else
        {
          MayDay::Error("AMRAdvectDiffuse::define: a_coarserLevelPtr is not castable to AMRAdvectDiffuse*");
        }
    }
  
  // Compute the grid spacing
  m_dx = m_domainLength / a_problemDomain.domainBox().longside();
  
  // Nominally, one layer of ghost cells is maintained permanently and
  // individual computations may create local data with more
  m_numGhost = 1;
  
  // Remove old patch integrator (if any), create a new one, and initialize
  if (m_patchGodunov != NULL)
    {
      delete m_patchGodunov;
    }
  
  m_patchGodunov = m_patchGodunovFactory->new_patchGodunov();
  m_patchGodunov->define(m_problem_domain,m_dx);
  
  // Get additional information from the patch integrator
  // note that m_numStates is number of advected scalars + Spacedim
  // since it also contains the advection velocities
  m_numStates = m_patchGodunov->numConserved();
  m_stateNames = m_patchGodunov->stateNames();

  // Set up physical BC object for diffusion. use HOExtrap here for 
  // lack of a better idea; eventually will probably want to be 
  // able to pass this in, then get from coarser levels.
  for (int dir=0; dir<SpaceDim; dir++)
    {
      SideIterator sit;
      for (sit.reset(); sit.ok(); ++sit)
        {
          HOExtrapBC thisBC(dir,sit());
          m_diffusionBC.setBoxGhostBC(thisBC);
        }
    }
  
}

// Advance by one timestep
Real AMRAdvectDiffuse::advance()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::advance level " << m_level << " to time " 
           << m_time << endl;
  }

  // Copy the new to the old
  m_UNew.copyTo(m_UNew.interval(),
                m_UOld,
                m_UOld.interval());

  Real newDt = 0.0;

  // Set up arguments to LevelGodunov::step based on whether there are
  // coarser and finer levels

  // Undefined flux register in case we need it
  LevelFluxRegister dummyFR;

  //Undefined leveldata in case we need it
  const LevelData<FArrayBox> dummyData;

  // containers for advective source term (u-dot-del-u) and diffusive piece
  // these need to have a ring of ghost cells
  IntVect ghostVect = IntVect::Unit;
  LevelData<FArrayBox> advectiveSrc(m_grids, m_numStates, ghostVect);
  // note that diffusiveSrc has 
  LevelData<FArrayBox> diffusiveSrc(m_grids, m_numStates, ghostVect);
  
#if 0 // this is obsolete, since m_UOld now contains velocities  
  //container for all state variables + advection velocity
  LevelData<FArrayBox> stateVect(m_grids, m_numStates+SpaceDim, ghostVect);
#endif
  
  // Set arguments to dummy values and then fix if real values are available
  LevelFluxRegister* coarserFRptr = NULL;
  LevelFluxRegister* finerFRptr   = NULL;

  const LevelData<FArrayBox>* coarserDataOldPtr = NULL;
  const LevelData<FArrayBox>* coarserDataNewPtr = NULL;
    
  Real tCoarserOld = 0.0;
  Real tCoarserNew = 0.0;

  // A coarser level exists
  if (m_hasCoarser)
  {
    AMRAdvectDiffuse* coarserPtr = getCoarserLevel();

    // Recall that my flux register goes between my level and the next
    // finer level
    coarserFRptr = &coarserPtr->m_fluxRegister;

    coarserDataOldPtr = &coarserPtr->m_UOld;
    coarserDataNewPtr = &coarserPtr->m_UNew;

    tCoarserNew = coarserPtr->m_time;
    tCoarserOld = tCoarserNew - coarserPtr->m_dt;
  }

  // A finer level exists
  if (m_hasFiner)
  {
    // Recall that my flux register goes between my level and the next
    // finer level
    finerFRptr = &m_fluxRegister;
  }

  // Advance the solve one timestep

  // first do hyperbolic part (compute vel-dot-del-phi
  // alternative would be to derive an advection diffusion levelGodunov
  // from the levelGodunov class and do the diffusion in there. 

#if 0
  // note that levelGodunov expects old-time solution to be in the 
  // advectiveSrc container for the update.
  m_UOld.copyTo(m_UOld.interval(), 
                stateVect, m_UOld.interval());
  
  Interval velComps(m_numStates, m_numStates+SpaceDim-1);
  fillAdvectionVelocities(stateVect, velComps);
#endif
  Interval velComps(m_numStates-SpaceDim, m_numStates-1);
  fillAdvectionVelocities(m_UOld, velComps);


  // for tracing step, diffusive source term is needed. this is a 
  // clunky way to do this, but can't think of another way to do this 
  // offhand, given that all we have is a HelmholtzOp.  Could do this 
  // with a PoissonOp, but want to keep things consistent, to where there
  // is only one levelop-derived class floating around
  m_helmop.applyOpI(m_UOld, coarserDataOldPtr, diffusiveSrc);
  // now need to back out diffusive src term for advection by subracting
  // UOld.
  DataIterator dit = m_UOld.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      diffusiveSrc[dit()] -= m_UOld[dit()];
      diffusiveSrc[dit()].negate();
      // for debugging
      //diffusiveSrc[dit()].setVal(0.0);
    }
  
  // fill in ghost cells for source
  DomainGhostBC diffusiveSrcBC;
  SideIterator sit;
  for (int dir=0; dir<SpaceDim; ++dir)
    {
      for (sit.reset(); sit.ok(); ++sit)
        {
          Extrap0thOrderBC thisBC(dir,sit());
          diffusiveSrcBC.setBoxGhostBC(thisBC);
        }
    }

  for (dit.begin(); dit.ok(); ++dit)
    {
      diffusiveSrcBC.applyInhomogeneousBCs(diffusiveSrc[dit()],
                                            m_problem_domain, m_dx);
    }

  // finally, need to do exchange 
  diffusiveSrc.exchange(diffusiveSrc.interval());

  // Since levelGodunov takes coarser data as a reference instead of a 
  // pointer, need to this little dance here
  if (!m_hasCoarser)
    {
      coarserDataOldPtr = &dummyData;
      coarserDataNewPtr = &dummyData;
      coarserFRptr = &dummyFR;
    }
  
  if (!m_hasFiner)
    {
      finerFRptr = &dummyFR;
    }


  newDt = m_levelGodunov.step(m_UNew,
                              *finerFRptr,
                              *coarserFRptr,
                              diffusiveSrc,
                              *coarserDataOldPtr,
                              tCoarserOld,
                              *coarserDataNewPtr,
                              tCoarserNew,
                              m_time,
                              m_dt);

  // Since levelGodunov takes coarser data as a reference instead of a 
  // pointer, need to this little dance here -- return to elliptic convention
  if (!m_hasCoarser)
    {
      coarserDataOldPtr = NULL;
      coarserDataNewPtr = NULL;
      coarserFRptr = NULL;
    }

  if (!m_hasFiner)
    {
      finerFRptr = NULL;
    }

  // since levelGodunov returns new-time solution instead of u-dot-del-phi,
  // need to subtract off uOld and then divide by dt
  // first copy from augmented state vector to advectiveSrc 
  m_UNew.copyTo(advectiveSrc.interval(), advectiveSrc, 
                advectiveSrc.interval());
  for (dit.begin(); dit.ok(); ++dit)
    {
      advectiveSrc[dit()] -= m_UOld[dit()];
      advectiveSrc[dit()] /= m_dt;
      // for debugging, set this to 0 (look at pure diffusion problem)
      //advectiveSrc[dit()].setVal(0.0);
    }
  // reset unew to uold if advection is turned off
  //m_UOld.copyTo(m_UOld.interval(), m_UNew, m_UNew.interval());

  // now compute diffusive piece of things
  m_diffusionSolverPtr->computeDiffusion(diffusiveSrc,
                                         m_UOld, 
                                         finerFRptr, 
                                         coarserFRptr, 
                                         coarserDataOldPtr,
                                         tCoarserOld,
                                         coarserDataNewPtr,
                                         tCoarserNew,
                                         advectiveSrc,
                                         m_time, m_dt);

  // now put everything together for update
  // (could also have had heatSolver do this using the updateSoln
  // function, but i wanted to test everything)
  // recall that UNew already contains UOld data along with the
  // advective piece of the update
  for (dit.begin(); dit.ok(); ++dit)
    {
      //m_UNew[dit()].plus(advectiveSrc[dit()], m_dt);
      m_UNew[dit()].plus(diffusiveSrc[dit()], m_dt);
    }
                                    
      

  // Update the time and store the new timestep
  m_time += m_dt;
  Real returnDt = m_cfl * newDt;

  m_dtNew = returnDt;

  return returnDt;
}

// Things to do after a timestep
void AMRAdvectDiffuse::postTimeStep()
{
  // Used for conservation tests
  static Real orig_integral = 0.0;
  static Real last_integral = 0.0;
  static bool first = true;

  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::postTimeStep " << m_level << endl;
  }

  if (m_hasFiner)
  {
    if (m_doImplicitReflux)
      {
        // first find out what time coarser level is (if it exists)
        Real crseTime = -1.0;
        if (m_level > 0) crseTime = m_coarser_level_ptr->time();

        // do multilevel operations if this is the coarsest level or if
        // coarser level is not at same time as this level. otherwise,
        // defer this until we get down to the coarsest level at this time.
        if (m_level == 0 || (abs(crseTime - m_time) > TIME_EPS))
          {
            int finest_level = m_level;
            // index through levels to find what finest level is
            AMRAdvectDiffuse* thisADPtr = this;
            while (thisADPtr->m_hasFiner)
              {
                thisADPtr = thisADPtr->getFinerLevel();
              }
            assert(!thisADPtr->m_hasFiner);
            finest_level = thisADPtr->m_level;

            // now do implicit refluxing
            Vector<LevelData<FArrayBox>* > refluxCorr(finest_level+1, NULL);
            Vector<LevelData<FArrayBox>* > refluxRHS(finest_level+1, NULL);
            Vector<DisjointBoxLayout> amrGrids(finest_level+1);
            Vector<int> amrRefRatios(finest_level+1,-1);
            Vector<Real> amrDx(finest_level+1);
            Vector<ProblemDomain> amrDomains(finest_level+1);

            // loop over levels, allocate storage, set up for AMRSolve
            thisADPtr = this;
            int startLev = m_level;
            // if coarser level exists, define it as well for BCs.
            if (startLev > 0) 
              {
                startLev -= 1;
                thisADPtr = thisADPtr->getCoarserLevel();
              }

            for (int lev = startLev; lev<= finest_level; lev++)
              {
                const DisjointBoxLayout& levelGrids = thisADPtr->m_grids;
                // rhs has no ghost cells, correction does
                refluxRHS[lev] = new LevelData<FArrayBox>(levelGrids,
                                                          m_numStates);

                refluxCorr[lev] = new LevelData<FArrayBox>(levelGrids,
                                                           m_numStates,
                                                           IntVect::Unit);
                amrGrids[lev] = levelGrids;
                amrRefRatios[lev] = thisADPtr->refRatio();
                amrDx[lev] = thisADPtr->m_dx;
                amrDomains[lev] = thisADPtr->problemDomain();

                // initialize corr, rhs to 0
                DataIterator levelDit = refluxRHS[lev]->dataIterator();
                LevelData<FArrayBox>& levelRHS = *refluxRHS[lev];
                LevelData<FArrayBox>& levelCorr = *refluxCorr[lev];
                for (levelDit.begin(); levelDit.ok(); ++levelDit)
                  {
                    levelRHS[levelDit()].setVal(0.0);
                    levelCorr[levelDit()].setVal(0.0);
                  }

                thisADPtr = thisADPtr->getFinerLevel();
              } // end loop over levels for setup.
            
            // now loop over levels and set up RHS, initial guess at soln
            // note that we don't need to look at finest level here,
            // since it will have no finer level to get a reflux correction
            // from.
            thisADPtr = this;
            for (int lev= m_level; lev<finest_level; lev++)
              {
                LevelData<FArrayBox>& levelRHS = *refluxRHS[lev];
                LevelData<FArrayBox>& levelCorr = *refluxCorr[lev];
                
                Real refluxScale = 1.0/amrDx[lev];
                thisADPtr->m_fluxRegister.reflux(levelRHS, refluxScale);

                // initial guess for correction is RHS
                levelRHS.copyTo(levelRHS.interval(), 
                                levelCorr, levelCorr.interval());

                thisADPtr->getFinerLevel();
              } // end loop over levels to compute RHS.

            // define levelop and solver for diffusive solves
            HelmholtzOp diffusionOp;
            
            diffusionOp.setDomainGhostBC(m_diffusionBC);
            Real beta = -m_nu*m_dt;
            diffusionOp.setHelmCoeff(beta);
            
            AMRSolver refluxSolver(amrGrids, amrDomains,
                                   amrDx, amrRefRatios, finest_level+1,
                                   m_level, &diffusionOp,
                                   m_numStates);
          
            refluxSolver.setTolerance(m_diffusion_solver_tol);
            
            if (s_verbosity >= 3) 
              {
                refluxSolver.setVerbose(true);
              }
            else
              {
                refluxSolver.setVerbose(false);
              }

#if NDEBUG
            // quick check -- want sum(RHS) to be 0
            for (int comp=0; comp<m_numStates; comp++) 
              {
                Interval thisInterval(comp,comp);
                Real sumRHS;
                sumRHS = computeSum(refluxRHS, amrRefRatios, m_dx,
                                    thisInterval, m_level);
                pout() << "Sum(RHS) for implicit reflux for "
                       << m_stateNames[comp] << " = " 
                       << setiosflags(ios::scientific) << sumRHS << endl;
              }
#endif

            // now do solve
            refluxSolver.solveAMR(refluxCorr, refluxRHS);
            
            // now increment solution with reflux correction
            // go from finest->coarsest so taht we can also do avgDown
            // increment AD pointer to finest level
            thisADPtr = this;
            while (thisADPtr->getFinerLevel() != NULL) 
              {
                thisADPtr = thisADPtr->getFinerLevel();
              }
            assert (thisADPtr->m_level == finest_level);
            for (int lev= finest_level; lev >= m_level; lev--)
              {
                LevelData<FArrayBox>& levelPhi = thisADPtr->m_UNew;
                LevelData<FArrayBox>& levelCorr = *(refluxCorr[lev]);
                DataIterator levelDit = levelCorr.dataIterator();
                for (levelDit.begin(); levelDit.ok(); ++levelDit)
                  {
                    levelPhi[levelDit()] += levelCorr[levelDit()];
                  }

                // if a finer level exists, do avgDown as well
                if (lev < finest_level)
                  {
                    // Average from finer level data
                    AMRAdvectDiffuse* fineAD = thisADPtr->getFinerLevel();
                    LevelData<FArrayBox>& fineU = fineAD->m_UNew;
                    
                    fineAD->m_coarseAverage.averageToCoarse(m_UNew,
                                                            fineU);
                  }
                // increment level pointer
                thisADPtr = thisADPtr->getCoarserLevel();
              } // end loop over levels for implicit refluxing
            
          
            // clean up storage
            for (int lev=0; lev<= finest_level; lev++)
              {
                if (refluxRHS[lev] != NULL)
                  {
                    delete refluxRHS[lev];
                  }
                if (refluxCorr[lev] != NULL) 
                  {
                    delete refluxCorr[lev];
                  }
              }
          } // end if a coarser level is not at the same time as this level
        
      }
    else
      {      
        // explicit Reflux
        Real scale = -1.0/m_dx; 
        m_fluxRegister.reflux(m_UNew,scale);
        
        // Average from finer level data
        AMRAdvectDiffuse* amrGodFinerPtr = getFinerLevel();
        
        amrGodFinerPtr->m_coarseAverage.averageToCoarse(m_UNew,
                                                        amrGodFinerPtr->m_UNew);
      } // end if we're doing explicit refluxing
  } // end if there is a finer level

  if (s_verbosity >= 2 && m_level == 0)
  {
    int nRefFine = 1;

    pout() << "AMRADvectDiffuse::postTimeStep:" << endl;
    pout() << "  Sums:" << endl;
    for (int comp = 0; comp < m_numStates; comp++)
    {
      Interval curComp(comp,comp);
      Real integral = computeSum(m_UNew,NULL,nRefFine,m_dx,curComp);

      pout() << "    " << setw(23)
                       << setprecision(16)
                       << setiosflags(ios::showpoint)
                       << setiosflags(ios::scientific)
                       << integral
             << " --- " << m_stateNames[comp];

      if (comp == 0 && !first) {
        pout() << " (" << setw(23)
                       << setprecision(16)
                       << setiosflags(ios::showpoint)
                       << setiosflags(ios::scientific)
                       << (integral-last_integral)
               << " " << setw(23)
                      << setprecision(16)
                      << setiosflags(ios::showpoint)
                      << setiosflags(ios::scientific)
                      << (integral-orig_integral)
               << ")";
      }

      pout() << endl;

      if (comp == 0)
      {
        if (first)
        {
          orig_integral = integral;
          first = false;
        }

        last_integral = integral;
      }
    }
  }

  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::postTimeStep " << m_level << " finished" << endl;
  }
}

// Create tags for regridding
void AMRAdvectDiffuse::tagCells(IntVectSet& a_tags) 
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::tagCells " << m_level << endl;
  }

  // Since tags are calculated using only current time step data, use
  // the same tagging function for initialization and for regridding.
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::tagCellsInit " << m_level << endl;
  }

  // Create tags based on undivided gradient of density
  IntVectSet localTags;
  const DisjointBoxLayout& levelDomain = m_UNew.disjointBoxLayout();
  // If there is a coarser level interpolate undefined ghost cells
  if (m_hasCoarser)
  {
    const AMRAdvectDiffuse* amrGodCoarserPtr = getCoarserLevel();

    PiecewiseLinearFillPatch pwl(levelDomain,
                                 amrGodCoarserPtr->m_UNew.disjointBoxLayout(),
                                 m_numStates,
                                 amrGodCoarserPtr->m_problem_domain,
                                 amrGodCoarserPtr->m_ref_ratio,
                                 1);

    pwl.fillInterp(m_UNew,
                   amrGodCoarserPtr->m_UNew,
                   amrGodCoarserPtr->m_UNew,
                   1.0,
                   0,
                   0,
                   m_numStates);
  }
  m_UNew.exchange(Interval(0,m_numStates-1));

  // Compute undivided gradient
  DataIterator dit = levelDomain.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
  {
    const Box& b = levelDomain[dit()];
    FArrayBox gradFab(b,SpaceDim);
    const FArrayBox& UFab = m_UNew[dit()];

    for (int dir = 0; dir < SpaceDim; ++dir)
    {
      const Box bCenter = b & grow(m_problem_domain,-BASISV(dir));
      const Box bLo     = b & adjCellLo(bCenter,dir);
      const int hasLo = ! bLo.isEmpty();
      const Box bHi     = b & adjCellHi(bCenter,dir);
      const int hasHi = ! bHi.isEmpty();
      FORT_GETGRAD(CHF_FRA1(gradFab,dir),
                   CHF_CONST_FRA1(UFab,0),
                   CHF_CONST_INT(dir),
                   CHF_BOX(bLo),
                   CHF_CONST_INT(hasLo),
                   CHF_BOX(bHi),
                   CHF_CONST_INT(hasHi),
                   CHF_BOX(bCenter));
    }

    FArrayBox gradMagFab(b,1);
    FORT_MAGNITUDE(CHF_FRA1(gradMagFab,0),
                   CHF_CONST_FRA(gradFab),
                   CHF_BOX(b));

    // Tag where gradient exceeds threshold
    BoxIterator bit(b);
    for (bit.begin(); bit.ok(); ++bit)
    {
      const IntVect& iv = bit();

      if (gradMagFab(iv) >= m_refineThresh)
      {
        localTags |= iv;
      }
    }
  }

  localTags.grow(m_tagBufferSize);

  // Need to do this in two steps unless a IntVectSet::operator &=
  // (ProblemDomain) operator is defined
  Box localTagsBox = localTags.minBox();
  localTagsBox &= m_problem_domain;
  localTags &= localTagsBox;

  Vector<IntVectSet> allTags;

  const int destProc = uniqueProc(SerialTask::compute);

  gather(allTags,localTags,destProc);

  if (procID() == uniqueProc(SerialTask::compute))
  {
    for (int i = 0; i < allTags.size(); ++i)
    {
      a_tags |= allTags[i];
    }
  }
}

// Create tags at initialization
void AMRAdvectDiffuse::tagCellsInit(IntVectSet& a_tags) 
{
  tagCells(a_tags);
}


// Set up data on this level after regridding
void AMRAdvectDiffuse::regrid(const Vector<Box>& a_newGrids)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::regrid " << m_level << endl;
  }

  // Save original grids and load balance
  m_level_grids = a_newGrids;
  m_grids = loadBalance(a_newGrids);

  if (s_verbosity >= 4)
  {
    // Indicate/guarantee that the indexing below is only for reading
    // otherwise an error/assertion failure occurs
    const DisjointBoxLayout& constGrids = m_grids;

    pout() << "new grids: " << endl;

    for (LayoutIterator lit = constGrids.layoutIterator(); lit.ok(); ++lit)
    {
      pout() << constGrids[lit()] << endl;
    }
  }

  // Save data for later
  LevelData<FArrayBox> UOld;
  UOld.define(m_UNew);
  m_UNew.copyTo(m_UNew.interval(),
                UOld,
                UOld.interval());

  // Reshape state with new grids
  IntVect ivGhost = m_numGhost*IntVect::Unit;
  m_UNew.define(m_grids,m_numStates,ivGhost);
  m_UOld.define(m_grids,m_numStates,ivGhost);

  // Set up data structures
  levelSetup();

  // Interpolate from coarser level
  if (m_hasCoarser)
  {
    AMRAdvectDiffuse* amrGodCoarserPtr = getCoarserLevel();
    m_fineInterp.interpToFine(m_UNew,amrGodCoarserPtr->m_UNew);
  }

  // Copy from old state
  UOld.copyTo(UOld.interval(),
                  m_UNew,
                  m_UNew.interval());
}

// Initialize grids
void AMRAdvectDiffuse::initialGrid(const Vector<Box>& a_newGrids)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::initialGrid " << m_level << endl;
  }

  // Save original grids and load balance
  m_level_grids = a_newGrids;
  m_grids = loadBalance(a_newGrids);

  if (s_verbosity >= 4)
  {
    // Indicate/guarantee that the indexing below is only for reading
    // otherwise an error/assertion failure occurs
    const DisjointBoxLayout& constGrids = m_grids;

    pout() << "new grids: " << endl;
    for (LayoutIterator lit = constGrids.layoutIterator(); lit.ok(); ++lit)
    {
      pout() << constGrids[lit()] << endl;
    }
  }

  // Define old and new state data structures
  IntVect ivGhost = m_numGhost*IntVect::Unit;
  m_UNew.define(m_grids,m_numStates,ivGhost);
  m_UOld.define(m_grids,m_numStates,ivGhost);

  // Set up data structures
  levelSetup();
}

// Initialize data
void AMRAdvectDiffuse::initialData()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::initialData " << m_level << endl;
  }

  PhysIBC* physIBCPtr = m_patchGodunov->getPhysIBC();
  physIBCPtr->initialize(m_UNew);

  // fill in advection velocities so that the initial plotfile won't
  // look weird
  Interval velComps(m_numStates-SpaceDim,m_numStates-1);
  fillAdvectionVelocities(m_UNew, velComps);

}

// Things to do after initialization
void AMRAdvectDiffuse::postInitialize()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::postInitialize " << m_level << endl;
  }

  // solve for post-regrid smoothing may eventually go here.
}

// Create a local PatchGodunov factory using the argument as a factory
void AMRAdvectDiffuse::patchGodunov(const PatchGodunov* const a_patchGodunov)
{
  if (m_patchGodunovFactory != NULL)
  {
    delete m_patchGodunovFactory;
  }

  m_patchGodunovFactory = a_patchGodunov->new_patchGodunov();
}


void
AMRAdvectDiffuse::diffusionSolver(const BaseHeatSolver* const a_heatSolverPtr)
{
  if (m_diffusionSolverPtr != NULL)
    {
      delete m_diffusionSolverPtr;
    }
  m_diffusionSolverPtr = a_heatSolverPtr->new_heatSolver();
}

#ifdef HDF5

// Write checkpoint header
void AMRAdvectDiffuse::writeCheckpointHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::writeCheckpointHeader" << endl;
  }

  // Setup the number of components
  HDF5HeaderData header;
  header.m_int["num_components"] = m_numStates;

  // Setup the component names
  char compStr[30];
  for (int comp = 0; comp < m_numStates; ++comp)
  {
    sprintf(compStr,"component_%d",comp);
    header.m_string[compStr] = m_stateNames[comp];
  }

  // Write the header
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }
}

// Write checkpoint data for this level
void AMRAdvectDiffuse::writeCheckpointLevel(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::writeCheckpointLevel" << endl;
  }

  // Setup the level string
  char levelStr[20];
  sprintf(levelStr,"%d",m_level);
  const std::string label = std::string("level_") + levelStr;

  a_handle.setGroup(label);

  // Setup the level header information
  HDF5HeaderData header;

  header.m_int ["ref_ratio"]       = m_ref_ratio;
  header.m_int ["tag_buffer_size"] = m_tagBufferSize;
  header.m_real["dx"]              = m_dx;
  header.m_real["dt"]              = m_dt;
  header.m_real["nu"]              = m_nu;
  header.m_box ["prob_domain"]     = m_problem_domain.domainBox();


  // Setup the periodicity info
  D_TERM(
         if (m_problem_domain.isPeriodic(0))
           header.m_int ["is_periodic_0"] = 1;
         else
           header.m_int ["is_periodic_0"] = 0; ,
    
         if (m_problem_domain.isPeriodic(1))
           header.m_int ["is_periodic_1"] = 1;
         else
           header.m_int ["is_periodic_1"] = 0; ,
         
         if (m_problem_domain.isPeriodic(2))
           header.m_int ["is_periodic_2"] = 1;
         else 
           header.m_int ["is_periodic_2"] = 0; );

  // Write the header for this level
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }

  // Write the data for this level
  write(a_handle,m_UNew.boxLayout());
  write(a_handle,m_UNew,"data");
}

// Read checkpoint header
void AMRAdvectDiffuse::readCheckpointHeader(HDF5Handle& a_handle)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::readCheckpointHeader" << endl;
  }

  // Reader the header
  HDF5HeaderData header;
  header.readFromFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << "hdf5 header data:" << endl;
    pout() << header << endl;
  }

  // Get the number of components
  if (header.m_int.find("num_components") == header.m_int.end())
  {
    MayDay::Error("AMRAdvectDiffuse::readCheckpointHeader: checkpoint file does not have num_components");
  }

  int numStates = header.m_int["num_components"];
  if (numStates != m_numStates)
  {
    MayDay::Error("AMRAdvectDiffuse::readCheckpointHeader: num_components in checkpoint file does not match solver");
  }

  // Get the component names
  std::string stateName;
  char compStr[60];
  for (int comp = 0; comp < m_numStates; ++comp)
  {
    sprintf(compStr,"component_%d",comp);
    if (header.m_string.find(compStr) == header.m_string.end())
    {
      MayDay::Error("AMRAdvectDiffuse::readCheckpointHeader: checkpoint file does not have enough component names");
    }

    stateName = header.m_string[compStr];
    if (stateName != m_stateNames[comp])
    {
      MayDay::Error("AMRAdvectDiffuse::readCheckpointHeader: state_name in checkpoint does not match solver");
    }
  }
}

// Read checkpoint data for this level
void AMRAdvectDiffuse::readCheckpointLevel(HDF5Handle& a_handle)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::readCheckpointLevel" << endl;
  }

  // Setup the level string
  char levelStr[20];
  sprintf(levelStr,"%d",m_level);
  const std::string label = std::string("level_") + levelStr;

  // Read the header for this level
  a_handle.setGroup(label);

  HDF5HeaderData header;
  header.readFromFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << "hdf5 header data:" << endl;
    pout() << header << endl;
  }

  // Get the refinement ratio
  if (header.m_int.find("ref_ratio") == header.m_int.end())
  {
    MayDay::Error("AMRAdvectDiffuse::readCheckpointLevel: file does not contain ref_ratio");
  }
  m_ref_ratio = header.m_int["ref_ratio"];

  if (s_verbosity >= 2)
  {
    pout() << "read ref_ratio = " << m_ref_ratio << endl;
  }

  // Get the tag buffer size
  if (header.m_int.find("tag_buffer_size") == header.m_int.end())
  {
    MayDay::Error("AMRAdvectDiffuse::readCheckpointLevel: file does not contain tag_buffer_size");
  }
  m_tagBufferSize=  header.m_int["tag_buffer_size"];

  if (s_verbosity >= 2)
  {
    pout() << "read tag_buffer_size = " << m_tagBufferSize << endl;
  }

  // Get dx
  if (header.m_real.find("dx") == header.m_real.end())
  {
    MayDay::Error("AMRAdvectDiffuse::readCheckpointLevel: file does not contain dx");
  }
  m_dx = header.m_real["dx"];

  if (s_verbosity >= 2)
  {
    pout() << "read dx = " << m_dx << endl;
  }

  // Get dt
  if (header.m_real.find("dt") == header.m_real.end())
  {
    MayDay::Error("AMRAdvectDiffuse::readCheckpointLevel: file does not contain dt");
  }
  m_dt = header.m_real["dt"];

  if (s_verbosity >= 2)
  {
    pout() << "read dt = " << m_dt << endl;
  }

  // Get nu
  if (header.m_real.find("nu") == header.m_real.end())
  {
    MayDay::Error("AMRAdvectDiffuse::readCheckpointLevel: file does not contain nu");
  }
  m_nu = header.m_real["nu"];

  if (s_verbosity >= 2)
  {
    pout() << "read nu = " << m_nu << endl;
  }

  // Get the problem domain
  if (header.m_box.find("prob_domain") == header.m_box.end())
  {
    MayDay::Error("AMRAdvectDiffuse::readCheckpointLevel: file does not contain prob_domain");
  }

  Box domainBox = header.m_box["prob_domain"];
      
  // Get the periodicity info -- this is more complicated than it really
  // needs to be in order to preserve backward compatibility 
  bool isPeriodic[SpaceDim];
  D_TERM(if (!(header.m_int.find("is_periodic_0") == header.m_int.end()))
           isPeriodic[0] =  (header.m_int["is_periodic_0"] == 1);
         else
           isPeriodic[0] = false; ,

         if (!(header.m_int.find("is_periodic_1") == header.m_int.end()))
           isPeriodic[1] =  (header.m_int["is_periodic_1"] == 1);
         else
           isPeriodic[1] = false; ,

         if (!(header.m_int.find("is_periodic_2") == header.m_int.end()))
           isPeriodic[2] =  (header.m_int["is_periodic_2"] == 1);
         else
           isPeriodic[2] = false;);

  m_problem_domain = ProblemDomain(domainBox,isPeriodic);

  // Get the grids
  Vector<Box> grids;
  const int gridStatus = read(a_handle,grids);

  if (gridStatus != 0)
  {
    MayDay::Error("AMRAdvectDiffuse::readCheckpointLevel: file does not contain a Vector<Box>");
  }

  // Create level domain
  m_grids = loadBalance(grids);

  // Indicate/guarantee that the indexing below is only for reading
  // otherwise an error/assertion failure occurs
  const DisjointBoxLayout& constGrids = m_grids;

  LayoutIterator lit = constGrids.layoutIterator();
  for (lit.begin(); lit.ok(); ++lit)
  {
    const Box& b = constGrids[lit()];
    m_level_grids.push_back(b);
  }

  if (s_verbosity >= 4)
  {
    pout() << "read level domain: " << endl;
    LayoutIterator lit = m_grids.layoutIterator();
    for (lit.begin(); lit.ok(); ++lit)
    {
      const Box& b = m_grids[lit()];
      pout() << lit().intCode() << ": " << b << endl;
    }
    pout() << endl;
  }

  // Reshape state with new grids
  m_UNew.define(m_grids,m_numStates);
  const int dataStatus = read<FArrayBox>(a_handle,
                                         m_UNew,
                                         "data",
                                         m_grids);

  if (dataStatus != 0)
  {
    MayDay::Error("AMRAdvectDiffuse::readCheckpointLevel: file does not contain state data");
  }
  m_UOld.define(m_grids,m_numStates);

  // Set up data structures
  levelSetup();
}

// Write plotfile header
void AMRAdvectDiffuse::writePlotHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::writePlotHeader" << endl;
  }

  // Setup the number of components -- include space for error
  HDF5HeaderData header;
  header.m_int["num_components"] = m_numStates+1;

  // Setup the component names
  char compStr[30];
  for (int comp = 0; comp < m_numStates; ++comp)
  {
    sprintf(compStr,"component_%d",comp);
    header.m_string[compStr] = m_stateNames[comp];
  }

  // include error
  int comp= m_numStates;
  sprintf(compStr, "component_%d",comp);
  header.m_string[compStr] = "error";

  // Write the header
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }
}

// Write plotfile data for this level
void AMRAdvectDiffuse::writePlotLevel(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::writePlotLevel" << endl;
  }

  // Setup the level string
  char levelStr[20];
  sprintf(levelStr,"%d",m_level);
  const std::string label = std::string("level_") + levelStr;

  a_handle.setGroup(label);

  // Setup the level header information
  HDF5HeaderData header;

  header.m_int ["ref_ratio"]   = m_ref_ratio;
  header.m_real["dx"]          = m_dx;
  header.m_real["dt"]          = m_dt;
  header.m_box ["prob_domain"] = m_problem_domain.domainBox();

  // Write the header for this level
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }

  // Write the data for this level
  LevelData<FArrayBox> plotData(m_UNew.getBoxes(), m_numStates+1);
  
  // first copy data to plot data holder
  m_UNew.copyTo(m_UNew.interval(), plotData, m_UNew.interval());

  // now compute error, if possible
  // first compute exact solution and place in place of error
  int errStartComp = m_numStates;
  computeExact(plotData, errStartComp);

  // now subtract off the current solution
  DataIterator dit = plotData.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      FArrayBox& errFab = plotData[dit()];
      int scalComp = 0;
      errFab.minus(m_UNew[dit()], scalComp, errStartComp, m_numStates-SpaceDim);
    }

  write(a_handle,m_UNew.boxLayout());
  write(a_handle,plotData,"data");
}

#endif

// Returns the dt computed earlier for this level
Real AMRAdvectDiffuse::computeDt()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::computeDt " << m_level << endl;
  }

  Real newDt;
  newDt = m_dtNew;

  return newDt;
}

// Compute dt using initial data
Real AMRAdvectDiffuse::computeInitialDt()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::computeInitialDt " << m_level << endl;
  }

#if 0 // obsolete since m_UOld now contains velocities
  // need to create augmented state vector with velocities to 
  // compute maxWaveSpeed
  LevelData<FArrayBox> stateVect(m_grids, m_numStates+SpaceDim);
  m_UOld.copyTo(m_UOld.interval(),
                stateVect, m_UOld.interval());

  Interval velComps(m_numStates, m_numStates+SpaceDim-1);
  fillAdvectionVelocities(stateVect, velComps);
#endif    
                                  

  Real newDT = m_initial_dt_multiplier * m_dx
    / m_levelGodunov.getMaxWaveSpeed(m_UNew);

  return newDT;
}

// Set the CFL number
void AMRAdvectDiffuse::CFL(Real a_cfl)
{
  m_cfl = a_cfl;
}

// set diffusion coefficient nu
void
AMRAdvectDiffuse::diffusionCoeff(Real a_nu)
{
  m_nu = a_nu;
}

// set whether to do implicit refluxing
void
AMRAdvectDiffuse::doImplicitReflux(bool a_doImplicitReflux)
{
  m_doImplicitReflux = a_doImplicitReflux;
}

// set tolerance for diffusion solvers
void
AMRAdvectDiffuse::diffusionSolverTolerance(Real a_solverTol)
{
  m_diffusion_solver_tol = a_solverTol;
}

// set background velocity field
void
AMRAdvectDiffuse::backgroundVelocity(const RealVect a_velocity)
{
  m_backgroundVelocity = a_velocity;
}

// Set the physical dimension of the longest side of the domain
void AMRAdvectDiffuse::domainLength(Real a_domainLength)
{
  m_domainLength = a_domainLength;
}

// Set the refinement threshold
void AMRAdvectDiffuse::refinementThreshold(Real a_refineThresh)
{
  m_refineThresh = a_refineThresh;
}

// Set the tag buffer size
void AMRAdvectDiffuse::tagBufferSize(int a_tagBufferSize)
{
  m_tagBufferSize = a_tagBufferSize;
}

// Create a load-balanced DisjointBoxLayout from a collection of Boxes
DisjointBoxLayout AMRAdvectDiffuse::loadBalance(const Vector<Box>& a_grids)
{
  // Load balance and create boxlayout
  Vector<int> procMap;
  if (procID() == uniqueProc(SerialTask::compute))
  {
    LoadBalance(procMap,a_grids);
  }
  broadcast(procMap,uniqueProc(SerialTask::compute));

  if (s_verbosity >= 4)
  {
    pout() << "AMRAdvectDiffuse::loadBalance: procesor map: " << endl;
    for (int igrid = 0; igrid < a_grids.size(); ++igrid)
    {
      pout() << igrid << ": " << procMap[igrid] << "  " << endl;
    }
    pout() << endl;
  }

  DisjointBoxLayout dbl(a_grids,procMap,m_problem_domain);
  dbl.close();

  return dbl;
}

// Setup menagerie of data structures
void AMRAdvectDiffuse::levelSetup()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRAdvectDiffuse::levelSetup " << m_level << endl;
  }

  AMRAdvectDiffuse* amrADCoarserPtr = getCoarserLevel();
  AMRAdvectDiffuse* amrADFinerPtr   = getFinerLevel();

  m_hasCoarser = (amrADCoarserPtr != NULL);
  m_hasFiner   = (amrADFinerPtr   != NULL);

  int nRefCrse = -1;
  DisjointBoxLayout* crseGridsPtr = NULL;

  if (m_hasCoarser)
  {
    nRefCrse = m_coarser_level_ptr->refRatio();
    crseGridsPtr = &amrADCoarserPtr->m_grids;

    const DisjointBoxLayout& coarserLevelDomain = amrADCoarserPtr->m_grids;



    m_coarseAverage.define(m_grids,
                           m_numStates,
                           nRefCrse);

    m_fineInterp.define(m_grids,
                        m_numStates,
                        nRefCrse,
                        m_problem_domain);


    // Maintain levelGodunov
    m_levelGodunov.define(m_grids,
                          coarserLevelDomain,
                          m_problem_domain,
                          nRefCrse,
                          m_dx,
                          m_patchGodunovFactory,
                          m_hasCoarser,
                          m_hasFiner);

    // This may look twisted but you have to do this this way because the
    // coarser levels get setup before the finer levels so, since a flux
    // register lives between this level and the next FINER level, the finer
    // level has to do the setup because it is the only one with the
    // information at the time of construction.
    
    // Maintain flux registers
    amrADCoarserPtr->m_fluxRegister.define(m_grids,
                                           amrADCoarserPtr->m_grids,
                                           m_problem_domain,
                                           amrADCoarserPtr->m_ref_ratio,
                                           m_numStates);
    amrADCoarserPtr->m_fluxRegister.setToZero();
  }
  else
  {
    m_levelGodunov.define(m_grids,
                          DisjointBoxLayout(),
                          m_problem_domain,
                          m_ref_ratio,
                          m_dx,
                          m_patchGodunovFactory,
                          m_hasCoarser,
                          m_hasFiner);
  }

  // define heat solver
  assert (m_diffusionSolverPtr != NULL);

  m_helmop.setHelmCoeff(-m_nu);
  m_helmop.setDomainGhostBC(m_diffusionBC);
  bool homogeneousOnly = false;
  m_helmop.define(m_grids, crseGridsPtr, m_dx, nRefCrse, 
                  m_problem_domain, homogeneousOnly, m_numStates);

  m_diffusionSolverPtr->define(m_grids, crseGridsPtr,
                               m_problem_domain,
                               m_dx, nRefCrse,
                               &m_helmop, m_numStates);
  


    

}

// Get the next coarser level
AMRAdvectDiffuse* AMRAdvectDiffuse::getCoarserLevel() const
{
  AMRAdvectDiffuse* amrADCoarserPtr = NULL;

  if (m_coarser_level_ptr != NULL)
  {
    amrADCoarserPtr = dynamic_cast<AMRAdvectDiffuse*>(m_coarser_level_ptr);

    if (amrADCoarserPtr == NULL)
    {
      MayDay::Error("AMRAdvectDiffuse::getCoarserLevel: dynamic cast failed");
    }
  }

  return amrADCoarserPtr;
}

// Get the next finer level
AMRAdvectDiffuse* AMRAdvectDiffuse::getFinerLevel() const
{
  AMRAdvectDiffuse* amrADFinerPtr = NULL;

  if (m_finer_level_ptr != NULL)
  {
    amrADFinerPtr = dynamic_cast<AMRAdvectDiffuse*>(m_finer_level_ptr);

    if (amrADFinerPtr == NULL)
    {
      MayDay::Error("AMRAdvectDiffuse::getFinerLevel: dynamic cast failed");
    }
  }

  return amrADFinerPtr;
}

void
AMRAdvectDiffuse::fillAdvectionVelocities(LevelData<FArrayBox>& a_state,
                                          const Interval a_vel_interval)
{

  // sanity checks.
  // need to be able to fit all of velocity components in
  assert (a_vel_interval.size() == SpaceDim);
  assert (a_vel_interval.begin() >= 0);
  assert (a_vel_interval.end() < a_state.nComp());

  // first example, set vel to be one everywhere
  DataIterator dit = a_state.dataIterator();
  for (dit.begin(); dit.ok(); ++dit) 
    {
      FArrayBox& thisState = a_state[dit()];
      
      int velComp = a_vel_interval.begin();
      for (int dir=0; dir<SpaceDim; dir++, velComp++) 
        {
          Real thisVel = m_backgroundVelocity[dir];
          thisState.setVal(thisVel, velComp);
        }
    }
}
  
void 
AMRAdvectDiffuse::computeExact(LevelData<FArrayBox>& a_exact,
                               int a_startComp) const
{

  DataIterator dit = a_exact.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      FArrayBox& thisExact = a_exact[dit()];
      // iterate over scalar components (don't include velocity)
      for (int scalComp = 0; scalComp<m_numStates-SpaceDim; scalComp++)
        {
          int exactComp = a_startComp+scalComp;             
          FORT_ADVEXACT(CHF_FRA1(thisExact, exactComp),
                        CHF_CONST_REAL(m_time),
                        CHF_REALVECT(m_backgroundVelocity),
                        CHF_CONST_REAL(m_dx),
                        CHF_CONST_REAL(m_nu),
                        CHF_CONST_INT(scalComp));
        } // end loop over scalar components
    } // end loop over boxes
}

