/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef  RAMPIBC_H_
#define  RAMPIBC_H_

#include  <iostream>

#include "LevelData.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "RealVect.H"
#include "AMRIO.H"

#include "PhysIBC.H"
#include "SolidF_F.H"

/// Example of a wedge at zero attack angle
/**
   Parameters:
     a_gamma   - Gamma for polytropic, gamma-law gas
     a_alpha   - The angle from horizontal to the wedge edge
     a_ms      - Mach shock number
     a_xcorner - Position of the corner/tip of the wedge
     a_artvisc - Artificial viscosity coefficient

   This entire problem is rotated so that the wedge edge is horizontal
   in the problem domain.
 */
class RampIBC:public PhysIBC
{
public:
  /// Null constructor
  /**
   */
  RampIBC();
  
  /// Constructor which defines parameters used by Fortran routines
  /**
   */
  RampIBC(const Real& a_gamma,
          const Real& a_alpha,
          const Real& a_ms,
          const Real& a_xcorner,
          const Real& a_artvisc);

  /// Destructor
  /**
   */
  ~RampIBC() {};

  /// Sets parameters in a common block used by Fortran routines
  /**
   */
  void setFortranCommon(const Real& a_gamma,
                        const Real& a_alpha,
                        const Real& a_ms,
                        const Real& a_xcorner,
                        const Real& a_artvisc);

  /// Set the flag m_isFortranCommonSet to true
  /**
     Set the flag m_isFortranCommonSet to true so that new IBCs made with
     new_physIBC() will have this flag set without calling setFortranCommon()
     (this is a clumsy design and should be improved).
   */
  void setFortranCommonSet();

  /// Factory method - this object is its own factory
  /**
     Return a pointer to a new PhysIBC object with m_isDefined = false (i.e.,
     its define() must be called before it is used) and m_isFortranCommonSet
     set to value of m_isFortranCommonset in the current (factory) object.
   */
  PhysIBC *new_physIBC();

  /// Set boundary fluxes
  /**
   */
  void fluxBC(FArrayBox&            a_F,
              const FArrayBox&      a_W,
              const FArrayBox&      a_Wextrap,
              const int&            a_dir,
              const Side::LoHiSide& a_side,
              const Real&           a_time);
  
  /// Set boundary slopes
  /**
     The boundary slopes in a_dW are already set to one sided difference
     approximations.  If this function doesn't change them they will be
     used for the slopes at the boundaries.
   */
  void setBdrySlopes(FArrayBox&       a_dW,
                     const FArrayBox& a_W,
                     const int&       a_dir,
                     const Real&      a_time);
  
  /// Set up initial conditions
  /**
   */
  void initialize(LevelData<FArrayBox>& a_U);
  
  /// Adjust boundary fluxes to account for artificial viscosity
  /**
   */
  void artViscBC(FArrayBox&       a_F,
                 const FArrayBox& a_U,
                 const FArrayBox& a_divVel,
                 const int&       a_dir,
                 const Real&      a_time);
  
protected:
  // True if the Fortran common block has been set
  bool m_isFortranCommonSet;
};

#endif
