/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "CFIVS.H"
#include "LayoutIterator.H"
#include "DataIterator.H"

bool 
CFIVS::isDefined() const
{
  return(isdefined);
} 

const IntVectSet& 
CFIVS::getFineIVS() const 
{
  assert(isdefined);
  return fiinterp_ivs;
}

void  
CFIVS::setDefaultValues()
{
  isdefined = false;
  fiinterp_ivs.define();
}
CFIVS::CFIVS()
{
  setDefaultValues();
}

CFIVS::~CFIVS()
{
  setDefaultValues();
}

CFIVS::CFIVS(
             const Box& Domain, 
             const Box& box_in,
             const DisjointBoxLayout& Levboxes, 
             int Direction, Side::LoHiSide hiorlo)
{
  setDefaultValues();
  ProblemDomain probdomain(Domain);
  define(probdomain,  box_in, Levboxes, 
         Direction, hiorlo);
           
}


CFIVS::CFIVS(
             const ProblemDomain& Domain, 
             const Box& box_in,
             const DisjointBoxLayout& Levboxes, 
             int Direction, Side::LoHiSide hiorlo)
{
  setDefaultValues();
  define(Domain,  box_in, Levboxes, 
         Direction, hiorlo);
           
}

void
CFIVS::define(
              const Box& Domain, 
              const Box& box_in,
              const DisjointBoxLayout& fine_boxes, 
              int direction, 
              Side::LoHiSide hiorlo)
{
  ProblemDomain probdomain(Domain);
  define(probdomain, box_in, fine_boxes, direction, hiorlo);
}



void
CFIVS::define(
              const ProblemDomain& Domain, 
              const Box& box_in,
              const DisjointBoxLayout& fine_boxes, 
              int direction, 
              Side::LoHiSide hiorlo)
{
  isdefined = true;
  assert(direction >= 0);
  assert(direction < SpaceDim);
  assert(!Domain.isEmpty());
  assert(Domain.contains(box_in));
  assert(fine_boxes.checkPeriodic(Domain));
  assert((hiorlo == Side::Lo) || (hiorlo == Side::Hi));

  //shift direction
  int hilo = sign(hiorlo);
    
  //create fine stencil
  Box finebox = box_in;
  Box edgebox;
  assert((hilo ==1) || (hilo == -1));
  if(hilo == -1)
    {
      edgebox = adjCellLo(finebox,direction,1);
    }
  else
    {
      edgebox = adjCellHi(finebox,direction,1);
    }
  edgebox &= Domain;
  if(!edgebox.isEmpty())
    {
      fiinterp_ivs.define(edgebox);
      LayoutIterator lit = fine_boxes.layoutIterator();
      Box periodicTestBox(Domain.domainBox());
      if (Domain.isPeriodic()) 
        {
          for (int idir=0; idir<SpaceDim; idir++)
            {
              if (Domain.isPeriodic(idir)) 
                periodicTestBox.grow(idir,-1);
            }
        }
      for(lit.reset(); lit.ok(); ++lit)
        {
          fiinterp_ivs -= fine_boxes[lit()];
          // only do this IF we're periodic _and_ both boxes
          // adjoin the domain box boundary somewhere
          if (Domain.isPeriodic() && !periodicTestBox.contains(edgebox)
              && !periodicTestBox.contains(fine_boxes[lit()]))
            {
              ShiftIterator shiftIt = Domain.shiftIterator();
              IntVect shiftMult = Domain.domainBox().size();
              Box shiftedBox = fine_boxes[lit()];
              for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
                {
                  IntVect shiftVect(shiftMult*shiftIt());
                  shiftedBox.shift(shiftVect);
                  fiinterp_ivs -= shiftedBox;
                  shiftedBox.shift(-shiftVect);
                }
            }
        }
    }
}
