/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Fri, July 16, 1999

#ifndef _LevelMG_H_
#define _LevelMG_H_  

#include <iostream>
#include <cmath>
#include <assert.h>
#include <cstdlib>
#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "DisjointBoxLayout.H"
#include "FArrayBox.H"
#include "LevelData.H"
#include "CoarseAverage.H"
#include "LevelOp.H"
#include "ProblemDomain.H"

/// Multigrid solver on a level
/**
   Multigrid solver on a level.
   This class is to be considered internal 
   to AMRSolver and not a part of the Chombo API.
 */
class LevelMG 
{
public:

  ///
  bool isDefined() const;

  /// 
  LevelMG();

  /// 
  LevelMG(const DisjointBoxLayout& a_ba,
          const DisjointBoxLayout* a_baseBaPtr, 
          Real a_dxLevel, 
          int a_refRatio,
          const  Box& a_domain,
          int a_nCoarserLevels, 
          const LevelOp* const a_opin,
          int a_ncomp=1);

  /// 
  LevelMG(const DisjointBoxLayout& a_ba,
          const DisjointBoxLayout* a_baseBaPtr, 
          Real a_dxLevel, 
          int a_refRatio,
          const  ProblemDomain& a_domain,
          int a_nCoarserLevels, 
          const LevelOp* const a_opin,
          int a_ncomp=1);

  /// 
  void define(const DisjointBoxLayout& a_ba,
              const DisjointBoxLayout* a_baseBaPtr, 
              Real a_dxLevel, int a_refRatio,
              const Box& a_domain,
              int a_nCoarserLevels, 
              const LevelOp* const a_opin,
              int a_ncomp=1);


  /// 
  void define(const DisjointBoxLayout& a_ba,
              const DisjointBoxLayout* a_baseBaPtr, 
              Real a_dxLevel, int a_refRatio,
              const ProblemDomain& a_domain,
              int a_nCoarserLevels, 
              const LevelOp* const a_opin,
              int a_ncomp);

  /// Constructor for coarsened version of object.
  void define(
              const LevelMG& L,
              int refCoarse,
              const LevelOp* const opin_a
              );

  /// Constructor for coarsened version of object.
  LevelMG(
          const LevelMG& L,
          int refCoarse,
          const LevelOp* opin_a
          );


  ///
  ~LevelMG();

  ///
  void clear();


  ///
  /**: Invoke relaxation step. Default is pure MG V-cycle, suitable 
     for use in multilevel solver application; otherwise, use approximate 
     solver such as CG at bottom level. It is assumed that the problem has
     already been put in residual-correction form. In particular, only
     the homogeneous form of the physical and coarse-fine boundary
     conditions need be invoked.
  */
  void mgRelax(LevelData<FArrayBox> & a_soln,
               const LevelData<FArrayBox> & a_rhs,
               bool bottomsolveflag);

  ///
  void setnumBottomGSRB(int a_numBottomGSRB)
  {m_numBottomGSRB = a_numBottomGSRB;}

  ///
  void setnumSmoothUp(int a_numSmoothUp)
  {m_numSmoothUp = a_numSmoothUp;}

  ///
  void setnumSmoothDown(int a_numSmoothDown)
  {m_numSmoothDown = a_numSmoothDown;}


  // this is a dangerous access function that should not generally be used.
  LevelOp* levelOpPtr();

  // this is another access function that is kinda bad
  LevelMG* lCoarsePtr();

protected:

  void setDefaultValues();
  void clearMemory();
  bool m_isDefined;

  //these are owned by levelmg
  //
  LevelData<FArrayBox>  m_resid;

  //
  LevelData<FArrayBox>  m_crseResid;

  //
  LevelData<FArrayBox>  m_crseCorr;

  //
  DisjointBoxLayout m_ba;

  //
  DisjointBoxLayout m_baCoarsened;

  //
  const DisjointBoxLayout* m_baseBaPtr;

  //
  Real m_dxLevel;

  //
  ProblemDomain m_domain;

  //
  int m_nCoarserLevels;

  //  coarser level
  LevelMG* m_lCoarsePtr;

  //  das operator
  LevelOp* m_levelopPtr;

  //
  int m_numBottomGSRB;

  //
  int m_numSmoothUp;

  //
  int m_numSmoothDown;

  //
  int m_refToCoar;

  //
  CoarseAverage m_averageOp;

private:
  // correct fine on intersection with crse
  //should only be called internally because this
  //is not written for  general LDF's
  void crseCorrect(LevelData<FArrayBox>& a_fine,
                   const LevelData<FArrayBox>& a_crse,
                   int a_refRat);
  void operator=(const LevelMG& levmgin){};
  LevelMG(const LevelMG& levmgin) {};
};


#endif


