/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Sat, July 18, 1999

#include "LevelMG.H"
#include "LevelMGF_F.H"

void
LevelMG::setDefaultValues()
{
  // set defaults
  m_numBottomGSRB = 16;
  m_numSmoothUp = 4;
  m_numSmoothDown = 4;
  m_lCoarsePtr = NULL;
  m_levelopPtr = NULL;
  m_dxLevel = -1;
  m_isDefined = false;
}

bool 
LevelMG::isDefined() const
{
  return m_isDefined;
}

void 
LevelMG::clearMemory()
{
  if (m_lCoarsePtr != NULL) 
    {
      delete m_lCoarsePtr;
      m_lCoarsePtr = NULL;
    }
  if(m_levelopPtr != NULL)
    {
      delete m_levelopPtr;
      m_levelopPtr = NULL;
    }
  m_isDefined = false;
}
// default constructor
LevelMG::LevelMG()
{
  setDefaultValues();
}

/// Constructor with DisjointBoxLayout, number of coarser levels
//this calls the levelmg big define function
LevelMG::LevelMG(const DisjointBoxLayout& a_ba,
                 const DisjointBoxLayout* a_baseBaPtr, 
                 Real a_dxLevel, 
                 int a_refRatio,
                 const Box& a_domain,
                 int a_nCoarserLevels, 
                 const LevelOp* const a_opin,
                 int a_ncomp)
{
  setDefaultValues();
  ProblemDomain physdomain(a_domain);
  define(a_ba,a_baseBaPtr, a_dxLevel, a_refRatio, physdomain,
         a_nCoarserLevels, a_opin, a_ncomp); 
}

/// Constructor with DisjointBoxLayout, number of coarser levels
//this calls the levelmg big define function
LevelMG::LevelMG(const DisjointBoxLayout& a_ba,
                 const DisjointBoxLayout* a_baseBaPtr, 
                 Real a_dxLevel, 
                 int a_refRatio,
                 const ProblemDomain& a_domain,
                 int a_nCoarserLevels, 
                 const LevelOp* const a_opin,
                 int a_ncomp)
{
  setDefaultValues();
  define(a_ba,a_baseBaPtr, a_dxLevel, a_refRatio, a_domain, 
         a_nCoarserLevels, a_opin, a_ncomp); 
}

// corresponding define function
void
LevelMG::define(const DisjointBoxLayout& a_ba,
                const DisjointBoxLayout* a_baseBaPtr, 
                Real a_dxLevel, 
                int a_refRatio,
                const  Box& a_domain,
                int a_nCoarserLevels, 
                const LevelOp* const a_opin,
                int a_ncomp)
{
  ProblemDomain physdomain(a_domain);
  define(a_ba, a_baseBaPtr, a_dxLevel, a_refRatio, a_domain,
         a_nCoarserLevels, a_opin, a_ncomp);
}


// corresponding define function
void
LevelMG::define(const DisjointBoxLayout& a_ba,
                const DisjointBoxLayout* a_baseBaPtr, 
                Real a_dxLevel, 
                int a_refRatio,
                const  ProblemDomain& a_domain,
                int a_nCoarserLevels, 
                const LevelOp* const a_opin,
                int a_ncomp)
{
  assert(a_opin != NULL);
  assert(a_nCoarserLevels >= 0);
  assert(!a_domain.isEmpty());
  assert(a_ba.checkPeriodic(a_domain));

  //either there are no coarser levels 
  //or we have a legitimate ref ratio
  assert((a_refRatio > 0) || (a_baseBaPtr == NULL));
  assert(a_dxLevel > 0);

  clearMemory();

  m_ba = a_ba;
  m_baseBaPtr = a_baseBaPtr;

  m_dxLevel = a_dxLevel;
  m_domain  = a_domain;

  m_nCoarserLevels = a_nCoarserLevels;

  m_isDefined = true;
  m_resid.define(m_ba, a_ncomp, IntVect::TheZeroVector());
  m_levelopPtr = a_opin->new_levelop();
  //no need to do inhomgeneous CFInterp here
  //all levelmg operations (smooth, etc) use
  //only homogeneous cf interpolation
  bool homogeneousOnly = true;
  m_levelopPtr->define(m_ba,  m_baseBaPtr, 
                       m_dxLevel,  a_refRatio, 
                       m_domain, homogeneousOnly, a_ncomp);
  // recursively define MG hierarchy
  if (m_nCoarserLevels != 0) 
    {

      //this constructor is called at the finest mg level.
      //this means that it might need an averaging operator
      //but it does not need an interpolation operator
      m_refToCoar = 2;
      m_averageOp.define(m_ba, a_ncomp, m_refToCoar);
      if (m_baCoarsened.isClosed()) {
        m_baCoarsened = DisjointBoxLayout();
      }
      coarsen(m_baCoarsened, m_ba, m_refToCoar);
      m_crseCorr.define(m_baCoarsened,a_ncomp,IntVect::TheUnitVector());
      m_crseResid.define(m_baCoarsened,a_ncomp,IntVect::TheZeroVector());
      m_lCoarsePtr = new LevelMG(*this, m_refToCoar, a_opin);
      if(m_lCoarsePtr == NULL)
        {
          MayDay::Error("Out of Memory in LevelMG::define");
        }
    } 
}

// --------------------------------------------------------------
//  constructor for coarsened version of levelMG
// refcoarse is the ratio from coarsened thing to this thing
// --------------------------------------------------------------
LevelMG::LevelMG(const LevelMG& L, int refCoarse, const LevelOp* a_opin) 
{
  //set pointers to NULL so define does not try to delete them
  setDefaultValues();
  define(L, refCoarse, a_opin);
}
// --------------------------------------------------------------
//  constructor for coarsened version of levelMG
// refcoarse is the ratio from coarsened thing to this thing
// --------------------------------------------------------------
void 
LevelMG::define(const LevelMG& L, int refCoarse, const LevelOp* const a_opin) 
{
  assert(refCoarse == 2);
  assert(a_opin != NULL);
  clearMemory();
  m_isDefined = true;
  m_ba = L.m_baCoarsened;
  Real dxcoarse = refCoarse*L.m_dxLevel;
  int ncomp = L.m_resid.nComp();
  
  ProblemDomain dprobc = coarsen(L.m_domain, refCoarse);
  m_nCoarserLevels = L.m_nCoarserLevels - 1;
  assert(m_nCoarserLevels >= 0);

  m_baseBaPtr = L.m_baseBaPtr;
  m_dxLevel = dxcoarse;
  m_domain = dprobc;
  m_resid.define(m_ba, ncomp, IntVect::TheZeroVector());
  m_levelopPtr = L.m_levelopPtr->new_levelop();

  m_levelopPtr->define(L.m_levelopPtr, refCoarse);

  // recursively define MG hierarchy
  if (m_nCoarserLevels != 0) 
    {
      m_refToCoar = 2;
      m_averageOp.define(m_ba, ncomp, m_refToCoar);
      coarsen(m_baCoarsened, m_ba, m_refToCoar);
      m_crseCorr.define(m_baCoarsened,ncomp,IntVect::TheUnitVector());
      m_crseResid.define(m_baCoarsened,ncomp,IntVect::TheZeroVector());
      m_lCoarsePtr = new LevelMG(*this, m_refToCoar, a_opin);
      if(m_lCoarsePtr == NULL)
        {
          MayDay::Error("Out of Memory in LevelMG::define");
        }
    } 
  else 
    {
      m_lCoarsePtr = NULL;
    }
}

// ------------------------------------------------------------
// destructor
// ------------------------------------------------------------
LevelMG::~LevelMG()
{
  clearMemory();
}

// -------------------------------------------------------------
void
LevelMG::mgRelax(LevelData<FArrayBox>& a_soln,
                 const LevelData<FArrayBox>& a_rhs,
                 bool a_bottomSolveFlag)
{
  assert(isDefined());
  int ncomp = a_rhs.nComp();
  assert(a_soln.nComp() == ncomp);
  assert (ncomp == m_resid.nComp());
  assert( a_soln.ghostVect() == IntVect::TheUnitVector());

  // now, either coarsen or call bottom solver
  if (m_lCoarsePtr != NULL) 
    {
      // first, smooth on this level
      for (int iter=0; iter< m_numSmoothDown; iter++) 
        {
          m_levelopPtr->smooth(a_soln,a_rhs);
        }

      DataIterator ditCoar = m_crseCorr.dataIterator();
      for(ditCoar.reset(); ditCoar.ok(); ++ditCoar)
        m_crseCorr[ditCoar()].setVal(0.0);
      // define residual on this grid 
      // using homogeneous cf bcs
      m_levelopPtr->applyOpH(a_soln,m_resid);

      DataIterator dit = m_resid.dataIterator();
      for(dit.reset(); dit.ok(); ++dit)
        {
          m_resid[dit()] -= a_rhs[dit()];
          m_resid[dit()].negate();
        }

      // coarsen residual and physical boundary condition
      assert(m_averageOp.isDefined());
      m_averageOp.averageToCoarse(m_crseResid, m_resid);

      // relax recursively on coarsened level
      m_lCoarsePtr->mgRelax(m_crseCorr, m_crseResid, a_bottomSolveFlag);

      //interpolate correction back up and modify solution
      crseCorrect(a_soln, m_crseCorr, m_refToCoar);


      // smooth again on this level
      for (int iter=0; iter < m_numSmoothUp; iter++) 
        m_levelopPtr->smooth(a_soln,a_rhs);

    } 
  else if (m_domain.domainBox().numPts() == 1)
    {
      // grid is 1x1, just need to smooth
      // first, smooth on this level
      for (int iter=0; iter<m_numBottomGSRB; iter++) 
        {
          m_levelopPtr->smooth(a_soln,a_rhs);
        }
    } 
  else if (a_bottomSolveFlag) 
    {
      // first, smooth on this level
      for (int iter=0; iter< m_numBottomGSRB; iter++) 
        {
          m_levelopPtr->smooth(a_soln,a_rhs);
        }
      // use bottom smoother
      m_levelopPtr->bottomSmoother(a_soln,a_rhs);
    } 
  else
    {
      // smooth on this level
      // do numSmoothDown iterations
      for (int iter=0; iter< m_numSmoothDown; iter++) 
        {
          m_levelopPtr->smooth(a_soln,a_rhs);
        }
    }

}


void 
LevelMG::crseCorrect(LevelData<FArrayBox> & a_fine,
                     const LevelData<FArrayBox> & a_crse,
                     int a_refRat)
{

  DataIterator crseIt = a_crse.dataIterator();
  const DisjointBoxLayout& crseBoxes = a_crse.getBoxes();
  const DisjointBoxLayout& fineBoxes = a_fine.getBoxes();

  for(crseIt.reset(); crseIt.ok(); ++crseIt)
    {
      const Box& crseBox = crseBoxes[crseIt()];
      const FArrayBox& crseFab = a_crse[crseIt()];

      FArrayBox& fineFab = a_fine[crseIt()];
      Box fineBox = fineBoxes[crseIt()];
      fineBox.coarsen(a_refRat);
      fineBox &= crseBox;
      //c set refinement box
      Box nrefbox(IntVect::TheZeroVector(), 
                  (a_refRat-1)*IntVect::TheUnitVector());

      if(!fineBox.isEmpty())
        {
          FORT_INTERPMG(CHF_FRA(fineFab),
                        CHF_FRA(crseFab),
                        CHF_BOX(fineBox),
                        CHF_CONST_INT(a_refRat),
                        CHF_BOX(nrefbox));
         }
    }
}


LevelOp*
LevelMG::levelOpPtr()
{
  return m_levelopPtr;
}

LevelMG*
LevelMG::lCoarsePtr()
{
  return m_lCoarsePtr;
}
