/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Fri, July 23, 1999

#ifndef _LevelSolver_H_
#define _LevelSolver_H_

#include <cstdlib>
#include <iostream>
#include <assert.h>
#include <cmath>
#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "FArrayBox.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "LevelMG.H"
#include "ProblemDomain.H"

/// Solves elliptic equations on a level
/**
   LevelSolver Solves elliptic equations on a level
   using multigrid.
*/
class LevelSolver
{

public:
  /// 
  /**
     Default constructor leaves LevelSolver undefined.
  */
  LevelSolver();

  /// 
  virtual ~LevelSolver();

  ///
  /**
     Full constructor.  Calls full define function.\\
     {\bf Inputs:} \\
     a_grids: Grid on which solution will be computed.
     \\
     a_baseGrids: pointer to next coarser grids.  
     Should be == NULL if there is no coarser level.
     \\
     a_domain: domain of solution level
     \\
     a_dxLevel: grid spacing at solution level
     \\
     a_nRefCrse: refinement ratio between two levels.
     Not used if there is no finer level
     \\
     a_opin: operator L used in solution of Lphi=rho
  */
  LevelSolver(const DisjointBoxLayout& a_grids, 
              const DisjointBoxLayout* a_baseGrids, 
              const Box& a_domain, 
              Real a_dxLevel,  
              int a_nRefCrse,
              const LevelOp* const a_opin,
              int a_ncomp = 1);


  ///
  /**
     Full constructor.  Calls full define function.\\
     {\bf Inputs:} \\
     a_grids: Grid on which solution will be computed.
     \\
     a_baseGrids: pointer to next coarser grids.  
     Should be == NULL if there is no coarser level.
     \\
     a_domain: domain of solution level
     \\
     a_dxLevel: grid spacing at solution level
     \\
     a_nRefCrse: refinement ratio between two levels.
     Not used if there is no finer level
     \\
     a_opin: operator L used in solution of Lphi=rho
  */
  LevelSolver(const DisjointBoxLayout& a_grids, 
              const DisjointBoxLayout* a_baseGrids, 
              const ProblemDomain& a_domain, 
              Real a_dxLevel,  
              int a_nRefCrse,
              const LevelOp* const a_opin,
              int a_ncomp);


  ///   Full define function.
  /**
     Full define function.\\
     {\bf Inputs:} \\
     a_grids: Grid on which solution will be computed.
     \\
     a_baseGrids: pointer to next coarser grids.  
     Should be == NULL if there is no coarser level.
     \\
     a_domain: domain of solution level
     \\
     a_dxLevel: grid spacing at solution level
     \\
     a_nRefCrse: refinement ratio between two levels.
     Not used if there is no finer level
     \\
     a_opin: operator L used in solution of Lphi=rho
  */
  virtual void define(const DisjointBoxLayout& a_grids, 
                      const DisjointBoxLayout* a_baseGrids,
                      const Box& a_domain, 
                      Real a_dxLevel,
                      int a_nRefCrse,
                      const LevelOp* const a_opin,
                      int a_ncomp=1);


  ///   Full define function.
  /**
     Full define function.\\
     {\bf Inputs:} \\
     a_grids: Grid on which solution will be computed.
     \\
     a_baseGrids: pointer to next coarser grids.  
     Should be == NULL if there is no coarser level.
     \\
     a_domain: domain of solution level
     \\
     a_dxLevel: grid spacing at solution level
     \\
     a_nRefCrse: refinement ratio between two levels.
     Not used if there is no finer level
     \\
     a_opin: operator L used in solution of Lphi=rho
  */
  virtual void define(const DisjointBoxLayout& a_grids, 
                      const DisjointBoxLayout* a_baseGrids,
                      const ProblemDomain& a_domain, 
                      Real a_dxLevel,
                      int a_nRefCrse,
                      const LevelOp* const a_opin,
                      int a_ncomp = 1);


  /// 
  bool isDefined() const;

  ///
  void clearMemory();

  ///
  void setDefaultValues();

  ///
  /**
     do a level solve on this level using multigrid
     uses homogeneous bcs at coarse/fine interfaces \\
     {\bf Inputs:} \\
     phi: initial guess at solution. 
     rhs: right-hand side. \\
     {\bf Outputs:} \\
     phi: solution.
  */
  void levelSolveH(LevelData<FArrayBox>& phi, 
                   const LevelData<FArrayBox>& rhs,
                   bool initializePhiToZero=true);

  ///
  /**
     do a level solve on this level using multigrid
     uses inhomogeneous bcs at interfaces\\
     {\bf Inputs:} \\
     phi: initial guess at solution.
     phic: solution at next coarser level.  should be == NULL
     if there is no coarser level.
     rhs: right-hand side. \\
     {\bf Outputs:} \\
     phi: solution.
  */
  void levelSolve(LevelData<FArrayBox>& phi, 
                  const LevelData<FArrayBox>* phic, 
                  const LevelData<FArrayBox>& rhs,
                  bool initializePhiToZero=true);

  /// set solution tolerance
  void setTolerance(Real a_tolerance);

  /// 
  void setVerbose(bool a_verbose);

  /// set max number multigrid vcycles to use in solution
  void setMaxIter(int a_maxIter);
protected:

  // LevelMG level solver object to relax on this level
  LevelMG m_levelMG;

  // LevelMG level solver object to relax on this level
  LevelOp* m_levelOpPtr;

  // has this LevelSolver been defined yet?
  bool m_isDefined;

  // grids at this level
  DisjointBoxLayout  m_grids;

  // problem domain at this level
  ProblemDomain m_domain;

  // residual
  LevelData<FArrayBox> m_resid;

  // useful scratch space.
  LevelData<FArrayBox> m_scratch;

  // correction
  LevelData<FArrayBox> m_corr;

  //
  int m_nRefCrse;

  //
  Real m_dxLevel;

  //
  int m_maxIter;

  //
  Real m_tolerance;

  //
  bool m_verbose;

private:
  void operator = (const LevelSolver&){;}
  LevelSolver(const LevelSolver&){;}

};

#endif


