/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Fri, July 23, 1999

#include "LevelSolver.H"
#include "IntVectSet.H"
#include "parstream.H"
#include "LayoutIterator.H"

using std::cout;
using std::endl;

///////////////
bool
LevelSolver::isDefined() const
{
  return m_isDefined;
}
///////////////
void 
LevelSolver::setDefaultValues()
{
  m_levelOpPtr = NULL;
  m_isDefined = false;
  m_nRefCrse = -1;
  m_dxLevel = -1.0;
  m_maxIter = 33;
  m_tolerance = 1.0e-10;
#ifdef CH_USE_FLOAT
  m_tolerance = sqrt(m_tolerance);
#endif
  m_verbose = false;
}

void 
LevelSolver::setVerbose(bool a_verbose)
{
  m_verbose = a_verbose;
}
///////////////
void 
LevelSolver::setTolerance(Real a_tolerance)
{
  m_tolerance = a_tolerance;
}
///////////////
void 
LevelSolver::setMaxIter(int a_maxIter)
{
  m_maxIter = a_maxIter;
}
///////////////
// default constructor
LevelSolver::LevelSolver()
{
  setDefaultValues();
}

// destructor
LevelSolver::~LevelSolver()
{
  clearMemory();
}

void 
LevelSolver::define(const DisjointBoxLayout& a_grids, 
                    const DisjointBoxLayout* a_baseGrids, 
                    const Box& a_domain, 
                    Real a_dxLevel, int a_nRefCrse,
                    const LevelOp* const a_opin,
                    int a_ncomp)
{
  ProblemDomain physdomain(a_domain);
  define(a_grids, a_baseGrids, a_domain, a_dxLevel, a_nRefCrse,
         a_opin, a_ncomp);
}


void 
LevelSolver::define(const DisjointBoxLayout& a_grids, 
                    const DisjointBoxLayout* a_baseGrids, 
                    const ProblemDomain& a_domain, 
                    Real a_dxLevel, int a_nRefCrse,
                    const LevelOp* const a_opin,
                    int a_ncomp)
{
  clearMemory();
  m_grids = a_grids;
  m_domain = a_domain;
  m_nRefCrse = a_nRefCrse;
  m_dxLevel = a_dxLevel;
  m_resid.define(m_grids, a_ncomp, IntVect::TheZeroVector());
  m_scratch.define(m_grids, a_ncomp, IntVect::TheZeroVector());
  m_corr.define(m_grids, a_ncomp, IntVect::TheUnitVector());
  m_isDefined = true;

  m_levelOpPtr = a_opin->new_levelop();
  m_levelOpPtr->define(m_grids, a_baseGrids, m_dxLevel, 
                       m_nRefCrse, m_domain, false, a_ncomp);

  // coarsen boxes once to start, because coarsest level boxes
  // must also be coarsenable (for proper coarse-fine interpolation)
  // however, if grid merely covers entire physical domain, then 
  // want to coarsen all the way down

  int nCoarserLevels = -1;

  // why do it this way rather than as a dense IVS?
  IntVectSet ivsCheck(m_domain.domainBox());
  LayoutIterator lit = m_grids.layoutIterator();
  for(lit.reset(); lit.ok(); ++lit)
    ivsCheck -= m_grids.get(lit());
  if(ivsCheck.isEmpty()) 
    nCoarserLevels += 1;

  int refrattot = 2;
  bool getOut = false;
  while(!getOut)
    {
      bool addone = true;
      for(lit.reset(); lit.ok(); ++lit)
        {
          Box blocal = m_grids.get(lit());
          if(refine(coarsen(blocal,refrattot),refrattot) != blocal)
            addone = false;
          if(!addone) break;
        }
      if(addone)
        {
          nCoarserLevels += 1;
          refrattot *= 2;
        }
      else
        {
          getOut = true;
        }
    }
  nCoarserLevels = Max(nCoarserLevels, 0);

  m_levelMG.define(m_grids, a_baseGrids, m_dxLevel,
                   m_nRefCrse, m_domain, nCoarserLevels, a_opin, a_ncomp);
}

// complete constructor
LevelSolver::LevelSolver(const DisjointBoxLayout& a_grids, 
                         const DisjointBoxLayout* a_baseGrids,
                         const Box& a_domain, 
                         Real a_dxLevel, int a_nRefCrse,
                         const LevelOp* const a_opin, int a_ncomp)
{
  setDefaultValues();
  ProblemDomain physdomain(a_domain);
  define(a_grids,  a_baseGrids, physdomain, a_dxLevel, 
         a_nRefCrse, a_opin, a_ncomp);
}


// complete constructor
LevelSolver::LevelSolver(const DisjointBoxLayout& a_grids, 
                         const DisjointBoxLayout* a_baseGrids,
                         const ProblemDomain& a_domain, 
                         Real a_dxLevel, int a_nRefCrse,
                         const LevelOp* const a_opin,
                         int a_ncomp)
{
  setDefaultValues();
  define(a_grids,  a_baseGrids, a_domain, a_dxLevel, 
         a_nRefCrse, a_opin, a_ncomp);
}


// -----------------------------------------------------------------
// returns LevelSolver to a basically undefined state
void 
LevelSolver::clearMemory()
{
  if(m_levelOpPtr != NULL)
    delete m_levelOpPtr;
  m_levelOpPtr = NULL;
}


// solve on just this level, inhomogeneous bcs
void
LevelSolver::levelSolve(LevelData<FArrayBox>& a_phi, 
                        const LevelData<FArrayBox>* a_phic, 
                        const LevelData<FArrayBox>& a_rhs,
                        bool a_initializePhiToZero)
{
  assert(isDefined());
  assert(a_phi.nComp() == a_rhs.nComp());
  assert(a_phi.nComp() == m_resid.nComp());

  /* compute initial residual 
     -- can't use residual() fn because 
     don't want to zero out higher levels
     -- also note that we use LNf, 
     because this is a LEVEL solve
  */

  assert(m_levelOpPtr->isDefined());
  DataIterator dit = a_phi.dataIterator();
  if (a_initializePhiToZero) {
    //initial guess of phi is zero
    for(dit.reset(); dit.ok(); ++dit)
      a_phi[dit()].setVal(0.0);
  }

  m_levelOpPtr->applyOpI(a_phi, a_phic, m_resid);

  for(dit.reset(); dit.ok(); ++dit)
    {
      m_resid[dit()] -= a_rhs[dit()];
      m_resid[dit()].negate();
      m_corr[dit()].setVal(0.0);
    }

  Vector<Real> initRes(m_resid.nComp());

  bool done = true;
  
  for (int comp=0; comp<m_resid.nComp(); comp++) 
    {
      Interval thisInterval(comp,comp);
      initRes[comp] = norm(m_resid, thisInterval, 0);

      
      // if initRes = 0, don't bother solving
      if (initRes[comp] >  0.01*m_tolerance) done=false;
    }

  if (done) return;
      
  int iter = 0;
  while (!done) 
    {
      m_levelMG.mgRelax(m_corr, m_resid, true);

      m_levelOpPtr->applyOpH(m_corr, m_scratch);

      for(dit.reset(); dit.ok(); ++dit)
        {
          m_scratch[dit()] -= m_resid[dit()];
          m_scratch[dit()].negate();
        }

      iter++;

      done = true;
      for (int comp=0; comp<m_resid.nComp(); comp++) 
        {
          Real currentRes = 0.0;
          Interval thisInterval(comp,comp);

          currentRes =norm(m_scratch, thisInterval, 0);

          if(m_verbose)
            pout() << "iteration #= " << iter 
                   << ": Max(res) = " << currentRes 
                   << endl;
          
          done = done && (currentRes <= m_tolerance*initRes[comp] 
                          || iter > m_maxIter);
        } // end loop over components

    }  // end solver iterations

  // update phi
  for(dit.reset(); dit.ok(); ++dit)
    {
      a_phi[dit()] += m_corr[dit()];
    }
}

// solve on just this level, homogeneous bcs
void
LevelSolver::levelSolveH(LevelData<FArrayBox>& a_phi, 
                         const LevelData<FArrayBox>& a_rhs,
                         bool a_initializePhiToZero)
{
  assert(isDefined());
  assert(a_phi.nComp() == a_rhs.nComp());
  assert(a_phi.nComp() == m_resid.nComp());

  /** compute initial residual 
      -- can't use residual() fn because 
      don't want to zero out higher levels
      -- also note that we use LNf, 
      because this is a LEVEL solve
  **/

  assert(m_levelOpPtr->isDefined());
  DataIterator dit = a_phi.dataIterator();
  if (a_initializePhiToZero) {
    //initial guess of phi is zero
    for(dit.reset(); dit.ok(); ++dit)
      a_phi[dit()].setVal(0.0);
  }

  m_levelOpPtr->applyOpH(a_phi, m_resid);

  for (dit.reset(); dit.ok(); ++dit)
    {
      m_resid[dit()] -= a_rhs[dit()];
      m_resid[dit()].negate();
      m_corr[dit()].setVal(0.0);
    }

  Vector<Real> initRes(m_resid.nComp());
  
  bool done = true;

  for (int comp=0; comp<m_resid.nComp(); comp++)
    {
      Interval thisInterval(comp,comp);
      initRes[comp] = norm(m_resid, thisInterval, 0);

  // if initRes = 0, don't bother solving
      if (initRes[comp] > 0.01*m_tolerance) done = false;
    }
  
  if (done) return;

  int iter = 0;
  while (!done) 
    {
      iter++;

      // compute modified residual LofPhi = res - LNf(corr)
      // use homogeneous form of ApplyOp because want to use 
      // homogeneous CF BC's for correction
      m_levelMG.mgRelax(m_corr, m_resid, true);

      m_levelOpPtr->applyOpH(m_corr, m_scratch);

      for (dit.reset(); dit.ok(); ++dit)
        {
          m_scratch[dit()] -= m_resid[dit()];
          m_scratch[dit()].negate();
        }

      done = true;
      for (int comp=0; comp<m_scratch.nComp(); comp++)
        {
          Interval thisInterval(comp,comp);
          Real currentRes = norm(m_scratch, thisInterval, 0);

          if(m_verbose)    
            pout() << "iteration #= " << iter 
                   << ": Max(res) = " << currentRes 
                   << endl;
          
          done = done && (currentRes <= m_tolerance*initRes[comp] 
                          || iter > m_maxIter);
        } // end loop over components
      
    }  // end solver iterations
  
  // update phi
  for (dit.reset(); dit.ok(); ++dit)
    {
      a_phi[dit()] += m_corr[dit()];
    }
}
