/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _QuadCFInterp_H_
#define _QuadCFInterp_H_

#include <iostream>
#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "BaseFab.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "QuadCFStencil.H"
#include "ProblemDomain.H"

///Quadratic coarse-fine interpolation utility
/** 
    Class to fill ghost cells on coarse-fine
    interface using quadratic interpolation
*/
class QuadCFInterp
{
public:

  ///constructors, destructors, defines

  ///
  /**
     Full Constructor.  makes all coarse-fine
     information and sets internal variables 
     calls full define
  */
  QuadCFInterp(
               const DisjointBoxLayout& a_fineBoxes,
               const DisjointBoxLayout* a_coarBoxes,
               Real  a_dxFine, 
               int a_refRatio,
               int a_nComp,
               const Box& a_domf);

  ///
  /**
     Full Constructor.  makes all coarse-fine
     information and sets internal variables 
     calls full define
  */
  QuadCFInterp(
               const DisjointBoxLayout& a_fineBoxes,
               const DisjointBoxLayout* a_coarBoxes,
               Real  a_dxFine, 
               int a_refRatio,
               int a_nComp,
               const ProblemDomain& a_domf);

  ///
  /**
     Full define function.  makes all coarse-fine
     information and sets internal variables 
  */
  void define(
              const DisjointBoxLayout& a_fineBoxes,
              const DisjointBoxLayout* a_coarBoxes,
              Real  a_dxFine, 
              int a_refRatio,
              int a_nComp,
              const Box& a_domf);

  ///
  /**
     Full define function.  makes all coarse-fine
     information and sets internal variables 
  */
  void define(
              const DisjointBoxLayout& a_fineBoxes,
              const DisjointBoxLayout* a_coarBoxes,
              Real  a_dxFine, 
              int a_refRatio,
              int a_nComp,
              const ProblemDomain& a_domf);


  /// 
  /**
     Default constructor leaves QCFI undefined.
  */
  QuadCFInterp();

  /// 
  ~QuadCFInterp();

  /// 
  /**
     return QCFI to undefined state
  */
  void clear();

  /// 
  /**
     Coarse / Fine interpolation operator.
  */
  void coarseFineInterp(
                        LevelData<FArrayBox>& a_phif,
                        const LevelData<FArrayBox>& a_phic
                        );

  ///
  /**
     has full define function been called?  return true if so
  */
  bool isDefined() const;


protected:
  //boxes for coarse buffers
  BoxLayout m_loCoarBoxes[SpaceDim];
  BoxLayout m_hiCoarBoxes[SpaceDim];

  //fake level identifier.  
  int m_level;

  //number of components
  int m_nComp;

  //refinement ratio between levels
  int m_refRatio;

  //mesh spacing at fine level
  Real m_dxFine;

  //cfstencils on low-side faces
  //lives on fine input grid layout
  LayoutData<QuadCFStencil> m_loQCFS[SpaceDim];

  //cfstencils on high-side faces
  //lives on fine input grid layout
  LayoutData<QuadCFStencil> m_hiQCFS[SpaceDim];

  //buffer for coarse data at low edges of fine grids 
  BoxLayoutData<FArrayBox> m_hiCoarBuffer[SpaceDim];

  // copiers to copy data from inputCoarLayout to m_hiCoarBuffer
  Copier m_hiCoarCopiers[SpaceDim];

  //buffer for coarse data at high edges of fine grids 
  BoxLayoutData<FArrayBox> m_loCoarBuffer[SpaceDim];

  // copiers to copy data from inputCoarLayout to m_loCoarBuffer
  Copier m_loCoarCopiers[SpaceDim];

  DisjointBoxLayout m_inputFineLayout;
  DisjointBoxLayout m_inputCoarLayout;
  ///has full define function been called?
  bool m_isDefined;

protected:

  /* Coarse / Fine interpolation operator.
     using one sided differencing
     only at a particular face
  */
  void coarseFineInterp(BaseFab<Real> & a_phif,
                        const BaseFab<Real> & a_phic,
                        const QuadCFStencil& a_qcfs,
                        const Side::LoHiSide a_hiorlo,
                        const int a_idir,
                        const Interval& a_variables) const;

  //get extended phi (lives next to interpivs)
  void getPhiStar(BaseFab<Real> & a_phistar, 
                  const BaseFab<Real> & a_phic,
                  const QuadCFStencil& a_qcfs,
                  const Side::LoHiSide a_hiorlo,
                  const int a_idir,
                  const Interval& a_variables) const;

  //interpolate over correct intvectset
  void interpOnIVS(BaseFab<Real> & a_phif,
                   const BaseFab<Real> & a_phiStar, 
                   const QuadCFStencil& a_qcfs,
                   const Side::LoHiSide a_hiorlo,
                   const int a_idir,
                   const Interval& a_variables) const;
};

#endif
