/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <fstream>
#include <iostream>

// base Mesh refinement class
// ---------
//  class MeshRefine
//
// Short Description:
// ------------------
//  Create new meshes based on tagged cells on a range of levels of a mesh
//  hierarchy.  Each level of tagged cells is used to generate a new mesh at
//  the next finer level.  The finest level in the output mesh will be one
//  level higher than the top of the range of levels given as input.  As a 
//  special case, use the same tags (appropriately refined) for all levels.
//  MeshRefine is intended as a pure virtual base class, from which should
//  be derived a class which supplied the actual box-generation algorithm
//  (for example, the BRMeshRefine class implements the Berger-Rigoutsos
//  algorithm).
//
// Usage:
// ------
//  Call the regrid functions after computing error estimates and tagging
//  cells.  To add a new mesh level, set \var{TopLevel} to the index of 
//  the finest level in the existing mesh and define tags on the finest
//   level.  To keep the existing number of mesh levels, set \var{TopLevel} 
//  to one less than the index of the finest level and don't define any tags 
//  on the finest level. If a single IntVectSet of tags is passed (instead 
//  of a Vector<IntVectSet>) then the same tags (properly refined) will be 
//  used for all the new meshes up to level \var{TopLevel}+1.  In any case, 
//  the meshes at levels \var{BaseLevel} and below are not modified.  The 
//  output argument \var{newmeshes} will be reallocated to the necessary 
//  size before being used.  When this function returns, the elements of the
//  \var{newmeshes} vector corresponding to the unchanged levels will be
//  filled in with copies of the levels from the old mesh vector.  The
//  variable \var{tags} is modified in an undefined way, so its contents
//  should not be relied upon.  The variable \var{BlockFactor} specifies a
//  minimum grid size.  Every grid box will have an integral
//  multiple of \var{RefRatio}*\var{ceiling(BlockFactor/RefRatio)} cells 
//  in each dimension and also lower index values that are integral multiples.
//
//
// Arguments:
// ----------
//  \var{newmeshes}                type: Vector<Vector<Box>>
//      (output, index[0:\var{TopLevel}+1]) vector containing the new mesh hierarchy
//
//  \var{tags}                     type: Vector<IntVectSet>
//      (input/output, index [0:\var{TopLevel}]) tagged cells to be covered by the
//      new meshes.  For any level L, the new mesh at level L+1 will try to cover
//      the tagged cells at level L.  NOTE: This variable is modified in an
//      undefined manner so its values should not be used on exit
// --or--
//  \var{tags}                     type: IntVectSet
//      (input/output) tagged cells on the \var{BaseLevel} mesh.  These tags
//      will be refined and reused for the other mesh levels.
//
//  \var{BaseLevel}                type: int
//      (input, range[0:\var{TopLevel}]) index of the finest mesh level that
//      will not be modified.
//
//  \var{TopLevel}                 type: int
//      (input, range[\var{BaseLevel}:\var{OldMeshes.size()-1]) index of the
//      finest mesh level that will be refined (i.e. the finest mesh level for
//      which tags are defined); the new mesh vector \var{newmeshes} will have
//      an element at index [\var{TopLevel}+1].
//
//  \var{Domains}                  type: Vector<ProblemDomain>
//      (input, index[0:\var{TopLevel}]) boundaries of the physical domain,
//      one mesh box at each level.  Must be consistent with \var{RefRatios}
//      (i.e. Domains[levelL+1] == Domains[levelL] * RefRatios[levelL])
//
//  \var{OldMeshes}                type: Vector<Vector<Box>>
//      (input, index[0:\var{TopLevel}]) the existing mesh hierarchy vector.
//      The \var{BaseLevel} boxes should be aligned on \var{BufferSize}-sized
//      internal boundaries.
//
//  \var{RefRatios}                type: Vector<int>&
//      (input, index[0:\var{TopLevel}], range[2^i,i>0])
//      the number of cells (in each direction) in level L+1 for each cell in
//      level L.  Must be an integral power of 2.
//
//
//  \var{FillRatio}                type: Real
//      (input, range[0:1]) the minimum acceptable ratio
//      of filled cells in a box in the Berger-Rigoutsos algorithm in
//      \func{makeBoxes} (typical value = 0.75
//
//  \var{BlockFactor}              type: int
//      (input, range[1:) )
//      factor  that determines the minimum size of a box
//      and what index values a box corner may have (i.e. what alignment is 
//      has). Grids are guaranteed to be coarsenable by the BlockFactor.
//
//  \var{BufferSize}               type: int
//      (input, range[1:), default=1)
//      minimum number of level L mesh cells between the boundary of a level L+1
//      mesh and the nearest level L-1 mesh cell.  This is used to guarantee 
//      that stencils can be computed across coarse-fine boundaries.
//
//  \var{MaxBoxSize}               type: int
//      (input, range[2*BufferSize:), default=\infinity)
//      largest number of grid points in any dimension that will be generated
//      in an individual grid box.  Must be at least twice the BufferSize.
//      
// Returns:
// --------
//  The first two arguments (\var{tags}, \var{newmeshes}) are modified.
//
//  Expensive validations are done only when debugging is enabled
//   (i.e. the DEBUG make variable is "TRUE").
//   
//
// Usage Notes:
// ------------
//  All the input vectors should be defined with max index >= \var{TopLevel}.
//  They should have values for indices [\var{BaseLevel}:\var{TopLevel}].
//  (except for \var{OldMeshes}, which must be defined for all indices).  The
//  new mesh vector \var{newmeshes} will be redefined up to index
//  \var{TopLevel}+1.  \var{RefRatios} should be defined such that
//  \var{RefRatios}[L] is the value to use to refine the level L mesh to
//  produce the level L+1 mesh.  The \var{tags} vector is modified in an
//  undefined manner.  The output variable \var{newmeshes} may not be
//  completely defined if an exception occurs.  If you do not know what the
//  optional arguments do, it is probably safe to ignore them.
//  The BlockFactor can be used to force a minimum box size.
//
// Long Description:
// -----------------
//  MeshRefine replaces the meshes at levels \var{BaseLevel}+1 to
//  \var{TopLevel} and creates a new mesh at level \var{TopLevel}+1.  Levels
//  \var{BaseLevel} and below in the new mesh hierarchy are the same as in
//  the old one.  The new meshes are generated so that the tagged cells at
//  level L are refined and included in the new mesh at level L+1 (except
//  for tagged cells that aren`t properly nested in the \var{BaseLevel}
//  mesh).  If TopLevel<0, no new meshes are created and a warning code
//  is returned.
//
// Numerical Algorithm:
// --------------------
//  The tagged cells at level L are used to generate a new (refined) mesh at
//  level L+1.  The \var{BaseLevel} mesh is unchanged.  The new meshes must
//  satisfy the proper nesting requirements, i.e. each new mesh box must be
//  contained in the proper nesting domain (PND) for the next lower level.  The
//  PNDs of all levels are derived from the \var{BaseLevel} mesh (see
//  \func{makePNDs}).  The basic idea is that the proper nesting domain of a
//  level L mesh contains the interior cells of the level L-1 PND, except for
//  faces that are on the physical boundary.  The base level PND is defined as
//  the interiors of the base level mesh boxes, except for boundary faces
//  (again) and box face cells that are surrounded by cells of other boxes.
//
//  In the following graphic, the PND for two mesh boxes with BufferSize=1 is
//  shown.  
//
//        ###################################################
//        #     |   |   |   |   |   |                       #
//        #     |   | o | o | o |   |                       #
//        #     |   |   |   |   |   |                       #
//        #     +---+---+---+---+---%---+---+---+---+---+---#
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     |   | o | o | o |   %   |   |   |   |   |   #
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     +---+---+---+---+---%---+---+---+---+---+---#
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     |   | o | o | o | o % o | o | o | o | o | o #
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     +---+---+---+---+---%---+---+---+---+---+---#
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     |   |   |   |   |   %   | o | o | o | o | o #
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     +---+---+---+---+---%---+---+---+---+---+---#
//        #                         |   |   |   |   |   |   #
//        #                         |   | o | o | o | o | o #
//        #                         |   |   |   |   |   |   #
//        ###################################################
//
//  where:
//   #    are physical boundaries (non periodic)
//   | +  are cell boundaries of the two mesh boxes
//   %    are box boundaries shared between boxes
//   o    are cells in the proper nesting domain
//
//  The tagged cells outside the proper nesting domains are discarded by
//  intersecting the set of tagged cells at each level with the PND at that
//  level.  This ensures that all cells in the refined meshes are properly
//  nested.  The actual mesh generation starts at the top of the existing mesh
//  (\var{TopLevel}) and works down.  At each level a set of mesh boxes that
//  includes all the tagged cells at that level is generated (using
//  \func{makeBoxes}, which is provided in the derived class (e.g. 
//  BRMeshRefine).  The goal is to minimize the number of boxes 
//  and maximize the ratio of tagged to untagged cells in each box.  
//  The boxes are refined to the next higher level and saved as the new 
//  mesh (in \var{newmeshes}).  The unrefined boxes are coarsened and
//  projected down to the next lower level where a buffer of bufferSize cells
//  is added around all the boxes (the opposite of taking the interior 
//  cells that is done in building the PND) and added to the tagged cells 
//  at the coarser level.  This ensures that the next coarser level will 
//  have meshes that will properly nest the meshes just generated.  This 
//  is repeated until the BaseLevel+1 mesh is replaced.
//
//
// Implementation Method:
// ----------------------
//  There are 5 basic steps to perform.  Two functions (\func{makePNDs} and
//  \func{makeBoxes}) are called to handle the most complex steps and the code
//  in this function handles the rest.
//  Start
//   1) Compute the proper nesting domains (PNDs).  A \var{Vector<IntVectSet>}
//      is created and passed to \func{makePNDs}, along with the grid and
//      boundary data.  Only the part of the Vector that is active
//      [\var{BaseLevel:TopLevel}] is modified.  The other elements of the
//      vector are not touched.
//   2) Removing (clipping) the tag cells outside the PNDs is trivial:
//      intersect the two \var{IntVectSets} (PND and tag cells) at each level.
//      [Note: The PND data is not needed after this step.]
//  Loop for L from \var{TopLevel} downto \var{BaseLevel}
//   3) \func{makeBoxes} is called to find a set of boxes that cover the tag
//      cells at this level.  The output is stored in a temp \var{BoxArray}
//   4) The temp \var{BoxArray} is refined and the result is stored in the
//      output mesh vector at level L+1.
//   5) The temp \var{BoxArray} is coarsened to the next coarser level, then
//      expanded by the buffer cells in each direction (clipping at the 
//      physical boundaries) and then  unioned with the tags at level L-1.  
//      This step does not have to be done for the base level.  The resulting
//      tags are used to generate the mesh at the next coarser level.
//  EndLoop
//
// Implementation Notes:
// ---------------------
//  The output vector \var{newmeshes} is redefined to the necessary size
//  without necessarily reusing the existing memory space.
//
///////////////////////////////////////////////////////////////////////////////


// Include files:

#include "MeshRefine.H"
#include "MayDay.H"

MeshRefine::MeshRefine() : m_isDefined(false)
{
}

MeshRefine::~MeshRefine()
{
}

MeshRefine::MeshRefine(const Box& a_baseDomain,
                           const Vector<int>& a_refRatios,
                           const Real a_fillRatio,
                           const int a_blockFactor,
                           const int a_bufferSize,
                           const int a_maxSize)
{
  ProblemDomain crseDom(a_baseDomain);
  define(crseDom, a_refRatios, a_fillRatio, a_blockFactor,
         a_bufferSize, a_maxSize);
}


MeshRefine::MeshRefine(const ProblemDomain& a_baseDomain,
                           const Vector<int>& a_refRatios,
                           const Real a_fillRatio,
                           const int a_blockFactor,
                           const int a_bufferSize,
                           const int a_maxSize)
{
  define(a_baseDomain, a_refRatios, a_fillRatio, a_blockFactor,
         a_bufferSize, a_maxSize);
}

void
MeshRefine::define(const Box& a_baseDomain,
                     const Vector<int>& a_refRatios,
                     const Real a_fillRatio,
                     const int a_blockFactor,
                     const int a_bufferSize,
                     const int a_maxSize)
{
  ProblemDomain crseDom(a_baseDomain);
  define(crseDom, a_refRatios, a_fillRatio, a_blockFactor,
         a_bufferSize, a_maxSize);
}



void
MeshRefine::define(const ProblemDomain& a_baseDomain,
                     const Vector<int>& a_refRatios,
                     const Real a_fillRatio,
                     const int a_blockFactor,
                     const int a_bufferSize,
                     const int a_maxSize)
{
  int maxLevel = a_refRatios.size();
  m_nRefVect.resize(maxLevel);
  m_vectDomains.resize(maxLevel+1);
  m_level_blockfactors.resize(maxLevel);

  m_vectDomains[0] = a_baseDomain;
  for (int lev=0; lev<maxLevel; lev++) {
    m_nRefVect[lev] = a_refRatios[lev];
    m_vectDomains[lev+1] = refine(m_vectDomains[lev], a_refRatios[lev]);
  }
  
  // do some quick sanity checks
  assert( a_blockFactor >= 1 );                      
  assert( a_bufferSize >= 0 );                         
  assert( (a_maxSize >= 2*a_bufferSize) || (a_maxSize == 0) );
  assert( (a_blockFactor <= a_maxSize) || (a_maxSize == 0) );             
  assert( a_fillRatio > 0.0 || a_fillRatio <= 1.0 );
  
  m_fillRatio = a_fillRatio;
  m_blockFactor = a_blockFactor;
  m_bufferSize = a_bufferSize;
  m_maxSize = a_maxSize;

  computeLocalBlockFactors();

  m_isDefined = true;

}


//
// This regrid function takes a single IntVectSet of tags
//
int
MeshRefine::regrid(Vector<Vector<Box> >&   a_newmeshes,
                   const IntVectSet&             a_tags,
                   const int               a_baseLevel,
                   const int               a_topLevel,
                   const Vector<Vector<Box> >& a_oldMeshes) const
{
    //
    // Handle special cases
    //
    assert (a_topLevel >= 0 );
    assert ( a_baseLevel < (a_topLevel+1) && a_baseLevel >= 0 );

    //
    // Create a vector of tags from the single set of tags that is given
    Vector<IntVectSet> tags_vector( a_topLevel+1 ,a_tags );

    // Refine the levels above the base level
    for( int i = 1 ; i <= a_topLevel ; i++ )
      {
         tags_vector[i] = refine(tags_vector[i-1], m_nRefVect[i-1] ) ;
      }
    
    // run the version of refine that takes a vector of tags
    return regrid(a_newmeshes, tags_vector,
                  a_baseLevel, a_topLevel, a_oldMeshes);
}




//
// This version of refine takes a Vector of tags
//
int
MeshRefine::regrid(Vector<Vector<Box> >&   a_newmeshes,
                   const Vector<IntVectSet>&     a_tags,
                   const int               a_baseLevel,
                   const int               a_topLevel,
                   const Vector<Vector<Box> >& a_OldMeshes) const
{

  //
  // Validate arguments and handle special cases
  //
  assert (isDefined());
  
  assert ( a_topLevel >= 0 );
  assert ( a_baseLevel < (a_topLevel+1) && a_baseLevel >= 0 );
  assert ( a_OldMeshes.size() >= a_topLevel + 1 );
  assert (   m_vectDomains.size() >= a_topLevel + 1 );
  assert( m_nRefVect.size() >= a_topLevel + 1 );
  assert(      a_tags.size() >= a_topLevel + 1 );           

  Vector<IntVectSet> modifiedTags;

  //
  // set the top level to be the finest level which actually has tags
  bool found_top = false;
  int TopLevel = a_topLevel;
  int new_finest_level;
  // need this because .size() returns an unsigned int
  int isize = a_tags.size();
  for (int level = Min(a_topLevel, isize-1);
       !found_top &&(level >= a_baseLevel);
       --level) 
    {
      if (!a_tags[level].isEmpty()) 
        { 
          found_top = true;
          TopLevel = level;
        }
    }

  if (found_top) 
    {

      // copy tags to local storage so we can mess with them.
      modifiedTags.resize(a_tags.size());
      for (int ilev = a_baseLevel; ilev <= TopLevel; ilev++)
        {
          modifiedTags[ilev] = a_tags[ilev];
        }
  

      
      //
      // reinitialize the output array
      //
      a_newmeshes.resize( TopLevel+2 ) ;
      
      //
      // Generate new meshes if requested.
      //
      if( TopLevel+1 > a_baseLevel ) 
        {
          
          Box domaint = m_vectDomains[a_baseLevel].domainBox();
          Box testdom = coarsen(domaint, m_blockFactor);
          testdom.refine(m_blockFactor);
          if(domaint != testdom)
            MayDay::Error("MeshRefine:domain and Blocking Factor incompatible");
          // [NOTE: the following validations are expensive so only do them when
          //        debugging.]
          
#if ! defined( NDEBUG )
          // all existing meshes should be within the boundaries
          for( int i = a_baseLevel ; i <= TopLevel ; i++ ) {
            Box minbox = a_OldMeshes[i][0];
            for (int ibox = 0; ibox < a_OldMeshes[i].size(); ++ibox)
              {
                minbox.minBox(a_OldMeshes[i][ibox]);
              }
            assert(m_vectDomains[i].contains(minbox));
          }
#endif
          // all tagged cells must be within the existing meshes
          // make sure that each level of \var{Domains} is consistent with the 
          // previous level and the refinement ratio.
          for( int i = a_baseLevel ; i < TopLevel ; i++ ) 
            {
              if(m_vectDomains[i+1].domainBox() 
                 != refine(m_vectDomains[i].domainBox(), m_nRefVect[i])) 
                MayDay::Error("MeshRefine:domains and refratios incompatible");
            }
          //
          // coarsen old meshes, tags, problem domains by the appropriate
          //  blocking factor  
          Vector<Vector<Box> > OldMeshes = a_OldMeshes;
          Vector<ProblemDomain> Domains = m_vectDomains;
          for (int level = 0; level < modifiedTags.size (); ++level)
            {
              // first intersect tags with domains
              // (note that in periodic case, this can
              // result in wrapping of tags that are 
              // outside the base domain box
              modifiedTags[level] &= m_vectDomains[level];

              // coarsen blocking factor, rounding upwards
              // the goal of this is to coarsen everything down
              // to a level which is m_blockFactor coarser than
              // the new fine level we're going to generate.  By
              // generating grids at this fake level, we can then
              // refine up to the new level, and the blocking factor
              // will be automatically enforced.
              modifiedTags[level].coarsen (m_level_blockfactors[level]);
            }
          for (int level = 0; level < OldMeshes.size (); ++level)
            {
              if (level < m_level_blockfactors.size())
                {
                  for (int box = 0; box < OldMeshes[level].size (); ++box)
                    {
                      OldMeshes[level][box].coarsen(m_level_blockfactors[level]);
                    }
                  Domains[level].coarsen(m_level_blockfactors[level]);
                }
            }
          //
          // Compute proper nesting domains from a_baseLevel up to TopLevel.
          // [NOTE: this leaves pnds[0:a_baseLevel-1] empty.]
          //dtg
          Vector<IntVectSet> pnds( TopLevel+1 ) ;
          IntVectSet oldmesh_ivs;
          for (int box = 0; box < OldMeshes[a_baseLevel].size(); ++box)
            {
              oldmesh_ivs |= OldMeshes[a_baseLevel][box];
            }
          
          // GHM 7/30/01
          // the buffer size argument to makePND needs to be coarsened by
          // the BlockFactor, since all the other arguments have been so
          // coarsened.  If not, a circumstance may occur where a set of
          // grids defined with a certain set of tags does not "refine" to
          // itself given the exact same set of tags.
          
          Vector<int> blocked_BufferSize(TopLevel+1);
          for( int level=0; level<=TopLevel; level++ ) {
            blocked_BufferSize[level] = 
              (m_bufferSize +m_level_blockfactors[level] - 1 )/
              m_level_blockfactors[level];
          }
          
          // note that this is called _after_ we do all the coarsening to 
          // enforce blocking factor!
          makePNDs(pnds, a_baseLevel, TopLevel, Domains, oldmesh_ivs,
                   blocked_BufferSize);
          
          
          //
          // Clip the tagged cells to the proper nesting domains by
          // intersecting the two sets.
          //
          for( int lvl = a_baseLevel ; lvl <= TopLevel ; lvl++ ) {
            modifiedTags[lvl] &= pnds[lvl] ;
          }
          
          //
          // Generate new meshes.
          //
          
          // At each level, starting at the top, generate boxes that cover the tag
          // cells using makeBoxes(), refine these boxes and save them in the
          // output \var{a_newmeshes}.  Take the unrefined boxes, add a buffer zone 
          // around each box, coarsen, and union with the tag cells on the next 
          // coarser level.  This modifies the tags variable.  To handle 
          // \var{BlockFactor}, coarsen everything before making the new
          // meshes and then refine the resulting mesh boxes.
          Vector<Box> lvlboxes ;  // new boxes on this level
          for( int lvl = TopLevel ; lvl >= a_baseLevel ; lvl-- ) {
            ProblemDomain lvldomain ; // domain of this level
            
            lvldomain = Domains[lvl] ; 
            
            // Union the meshes from the previous level with the tags on this
            // level to guarantee that proper nesting is satisfied.  On the
            // first iteration this is a no-op because \var{lvlboxes} is empty.
            // [NOTE: for every iteration after the first, \var{lvlboxes} will
            //        already be coarsened by \var{BlockFactor} so it will be
            //        at the same refinement level as \var{tags[lvl]}, which 
            //        has also been coarsened]
            // this is simple in the non-periodic case, more complicated 
            // in the periodic case 
            if (lvldomain.isPeriodic() )
              {
                const Box domainBox = lvldomain.domainBox();
                ShiftIterator shiftIt = lvldomain.shiftIterator();
                IntVect shiftMult(domainBox.size());
                for (int i=0; i<lvlboxes.size(); i++)
                  {
                    Box localBox(lvlboxes[i]);
                    // will handle periodic wraparound through shifting
                    // and adding shifted image to tags, which will all
                    // remain within the domainBox
                    localBox &= domainBox;
                    modifiedTags[lvl] |= localBox;
                    // now do shifts to capture periodic images 
                    // necessary to enforce proper nesting
                    // only do this if original box was not 
                    // contained by domainBox
                    if (localBox != lvlboxes[i]) 
                      {
                        for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
                          {
                            IntVect shiftVect(shiftIt()*shiftMult);
                            Box localShiftedBox(lvlboxes[i]);
                            localShiftedBox.shift(shiftVect);
                            localShiftedBox &= domainBox;
                            if (!localShiftedBox.isEmpty())
                              modifiedTags[lvl] |= localShiftedBox;
                          } // end loop over shift directions
                      } // end whether periodic checking was needed
                  } // end loop over finer-level boxes to enforce nesting 
              } 
            else 
              {
                // non periodic case is simple
                for( int i = 0 ; i < lvlboxes.size() ; i++ )
                  modifiedTags[lvl] |= lvlboxes[i] ;
              }
            
            // make a new mesh at the same level as the tags
            makeBoxes(lvlboxes, modifiedTags[lvl], pnds[lvl], 
                      lvldomain, m_maxSize/m_blockFactor);
            
            
            // refine the new mesh and save it
            //[NOTE: we have to undo the coarsening by \var{BlockFactor} as well
            //       as refine to the next level.]
            //[NOTE: refine() operates in-place so copy first then refine()
            //       because the unrefined mesh will be needed later.]
            a_newmeshes[lvl+1] = lvlboxes ;
            for (int ibox = 0; ibox < a_newmeshes[lvl+1].size(); ++ibox)
              {
                a_newmeshes[lvl+1][ibox].refine(m_level_blockfactors[lvl]*m_nRefVect[lvl]);
              }
            // Make the boxes ready for the next iteration.
            // Don't have to do this for the last iteration.
            if( lvl > a_baseLevel ) 
              {
                
                // coarsen the unrefined new mesh so it matches the tags
                // at that level, then add the buffer cells, clip at the domain
                // boundaries so we can union them in the next iteration
                for( int i = 0 ; i < lvlboxes.size() ; i++ ) 
                  {
                    // since we want to have a BufferSize border
                    // of (lvl-1) cells, this translates into 
                    // growing the coarsened boxes by the BufferSize
                    // coarsened by the same factor we used to get the
                    // tags from level (lvl-1) to this intermediate level
                    lvlboxes[i].grow(Max(1, m_bufferSize/m_level_blockfactors[lvl]));
                    lvlboxes[i] &= Domains[lvl];
                    
                    // (dfm 3/26/2001) since you're working on a set of 
                    // faux levels which are a factor of m_blockFactor
                    // coarser than the levels you are _GENERATING_
                    // (in this case lvl+1), when you coarsen these
                    // boxes down to the next coarser level, you
                    // want to coarsen them by the refinement 
                    // ratio between levels (lvl+1) and (lvl),
                    // instead of the ratio between (lvl) and
                    // (lvl-1).
                    lvlboxes[i].coarsen( m_nRefVect[lvl] ) ;
                  }
                
              }
          }
        }
      
      //
      // Finally, copy the old mesh levels that didn't change.
      //
      for( int i = 0 ; i <= a_baseLevel ; i++ )
        a_newmeshes[i] = a_OldMeshes[i] ;
      
      //
      // Done generating grids
      //
      
      // set new finest level
      new_finest_level = TopLevel+1;
      // this is designed to catch the pathological but possible case
      // where there were tags on the TopLevel, but no grids were generated
      // (possibly if all tags were outside pnds)
      while (a_newmeshes[new_finest_level].size() == 0 
             && new_finest_level > a_baseLevel) 
        {
          new_finest_level -= 1;
        }

    } else {
      // if no tags on any level, just return
      new_finest_level = a_baseLevel;
    }
  
  return new_finest_level;
  
  
}// end of regrid function


void
MeshRefine::computeLocalBlockFactors()
{
  // computes amount that tags on a level l must be coarsened so
  // that grids on next finer level l+1 will be guaranteed to 
  // satisfy the blocking factor.
  // The easiest way to do this is to coarsen everything down
  // to a level which is m_blockFactor coarser than the new 
  // fine level which will be generated.  So, m_level_blockfactor[lvl]
  // is the amount by which the tags, etc at level lvl must
  // be coarsened so that they are m_blockFactor coarser than
  // the (lvl+1) grids which will be generated.  If m_BlockFactor
  // is less than the refinement ratio between levels (lvl) and (lvl+1),
  // then no coarsening needs to be done, so we default to one, in that case.
  for (int lev=0; lev<(m_level_blockfactors.size()); lev++)
    {
      m_level_blockfactors[lev] = (m_blockFactor + m_nRefVect[lev] -1)/m_nRefVect[lev];
    }
}

///////////////////////////////////////////////////////////////////////////////
//
// Function:
// ---------
//  int makePNDs()
//
// Short Description:
// ------------------
//  Compute the proper nesting domains for a subset of the levels in a
//  multilevel adaptive mesh.
//
// Usage:
// ------
//  This is called by \func{regrid}.  It wasn't designed to be called
//  directly so if you do so you're on your own.  The output is a
//  \var{Vector<IntVectSet>}.  This function is called to compute the proper
//  nesting domains needed to do mesh refinement.  The function that calls this
//  one must have the mesh and problem domain data structures already defined 
//  for all the levels and have allocated a vector<IntVectSet> for the proper 
//  nesting domains.  This function computes the proper nesting domains of the
//  levels to be refined (i.e., the levels \var{a_baseLevel} and above) and 
//  modifies the \var{pnds} array.  The output vector is undefined for 
//  indices corresponding to mesh levels below \var{a_baseLevel}
//  (i.e. [0:\var{a_baseLevel}-1]).
//
//  NOTE: in MeshRefine::regrid, this function is called _after_ everthing 
//        has been coarsened to enforce the blocking factor, so this behaves
//        differently than one might expect -- call outside of the regrid fn
//        at your own risk! (DFM 8/24/01)
//
// Arguments:
// ----------
//  \var{pnds}                     type: Vector<IntVectSet>
//      (output, index[\var{a_baseLevel}:\var{TopLevel}]) proper nesting domains
//
//  \var{a_baseLevel}                type: int
//      (input, range[0:\var{TopLevel}]) index of the mesh level to use as the
//      source for the proper nesting domains
//
//  \var{TopLevel}                 type: int
//      (input, range[\var{a_baseLevel}:\var{OldMeshes.size()-1]) index of the
//      finest mesh level to compute a PND for
//
//  \var{BaseMesh}                 type: Vector<Box>
//      (input) the boxes at level \var{a_baseLevel} in the mesh hierarchy.
//
//  \var{Domains}                  type: Vector<Box>
//      [same as for \func{MeshRefine}.]
//
//  \var{BufferSize}              type: int
//      [same as for \func{MeshRefine}.]
//
// Returns:
// --------
//  An int giving an exception number is returned.  The argument
//  \var{pnds} is modified.
//
// References:
// -----------
//  See ??? for a description of Proper Nesting Domains.
//  meshRefine -- calls this function
//
// Numerical Algorithm:
// --------------------
//  The proper nesting domain (PND) of level L+1 is defined as the interior of
//  the PND for level L (the next coarser) except at the boundaries of the
//  problem domain (ie, the level 0 box), in which case the PNDs of levels L
//  and L+1 both contain the boundary.  The PND of the base level is defined as
//  the interior of the union of all the grid boxes in the base level.  Given
//  the PND of level L (denoted by PND[L]) this function computes PND[L+1] by
//  iterating over the coordinate directions and computing the cells to remove
//  from the proper nesting domain due to the boundaries in that direction.
//  For each direction the domain is shifted (first up, then down) and only the
//  part that is shifted _out_ of the PND but remains _inside_ the problem
//  boundary is kept and then shifted back into the PND and subtracted
//  (removed) from it.  So if the shift moves part of the PND outside the
//  problem boundary then no cells will be removed.  If the cells that get
//  shifted out of the PND remain inside the boundary then they will be removed
//  from the PND when they get shifted back.  When all the directions have been
//  handled, the PND is refined to make it correct for level L+1.  The PND of
//  the base level is the union of the mesh boxes in the base level and serves
//  as the starting point.  In set notation the operation is:
//
//            D - ( ( ( (D << d) - D ) * B ) >> d )
//
//  where:
//    d is the current direction (+i,-i,+j,etc)
//    B is the boundary of the problem domain at the current level
//    - is the subtraction (removal) operator
//    * is the intersection operator
//   << is the down-shift operator (shift in negative direction)
//   >> is the up-shift operator (shift in positive direction)
//
//
// Implementation Method:
//  The PNDs are implemented as a vector of \type{IntVectSet}s.  A scratch 
//  IntVectSet is used to store the PND of the current level while it is 
//  being computed.  When the PND for a level is complete, it is refined 
//  and used for the next level.  The base PND is computed as the union of 
//  the mesh boxes in the base level mesh.  The PNDs for each level are 
//  stored in the appropriate elements of the pnds vector as they are 
//  computed.  Levels below a_baseLevel are not modified.  The code loops 
//  through each level starting from Baselevel.  Loop through each coordinate
//  direction applying the operation defined in the "Numerical Algorithm" 
//  section above.  Copy the final result into the PNDs vector.  If an 
//  exception occurs, do not change PNDs on the level that took the
//   exception.
//
// Implementation Notes:
//  This assumes cell-centers are being manipulated, not vertices.  The whole
//  vector of pnds is passed even though only some are accessed
//  because it is more convenient since these variables are likely to exist
//  already.
//
///////////////////////////////////////////////////////////////////////////////

void
MeshRefine::makePNDs(Vector<IntVectSet>&       a_pnds, 
                       const int            a_baseLevel,
                       const int             a_topLevel,
                       const Vector<ProblemDomain>&     a_domains,
                       const IntVectSet&    a_baseMesh, 
                       const Vector<int>&   a_bufferSize) const
{
  //
  // Validate inputs
  //
  assert( a_baseLevel <= a_topLevel && a_baseLevel >= 0 );
  assert(   a_domains.size() >= a_topLevel + 1 );       
  assert( m_nRefVect.size() >= a_topLevel + 1 );       
  // all existing boxes on the base level must be aligned to \var{BuffSize}


  a_pnds[a_baseLevel] = a_baseMesh;
  
  for( int lvl = a_baseLevel ; lvl <= a_topLevel ; lvl++)
    {
      IntVectSet& pnd = a_pnds[lvl];
      
      pnd.nestingRegion(a_bufferSize[lvl], a_domains[lvl]);

      if( (a_topLevel - lvl) > 0)
        {
          a_pnds[lvl+1] = pnd;


          // note that we use the refinement ratio btwn 
          // lvl+1 and lvl+2 in this case.  this is because
          // when this has been called from MeshRefine,
          // each level has already been brought the the
          // blocking-factor-enforcing coarseness of 
          // lvl+1 coarsened by the blocking factor
          // (since that's where the grids produced by
          // the tags on lvl will eventually wind up)
          a_pnds[lvl+1].refine(m_nRefVect[lvl+1]);
        }
      
    }
  
}
//end of makePNDs



// ---------------------------------------------------------------
// Utility and access functions
// --------------------------------------------------------------

  /// returns vector of refinement ratios
const Vector<int>& 
MeshRefine::refRatios() const
{
  return m_nRefVect;
}
  
/// returns fillRatio
Real 
MeshRefine::fillRatio() const
{
  return m_fillRatio;
}
  
/// returns blocking factor
int 
MeshRefine::blockFactor() const
{
  return m_blockFactor;
}
  
/// returns proper nesting buffer size 
int 
MeshRefine::bufferSize() const
{
  return m_bufferSize;
}
  
/// returns maximum box size in any dimension
int 
MeshRefine::maxSize() const
{
  return m_maxSize;
}

/// sets vector of refinement ratios
void 
MeshRefine::refRatios(const Vector<int>& a_nRefVect)
{
  int isize = a_nRefVect.size();
  m_nRefVect.resize(isize);
  for (int i=0; i<isize; i++) 
    {
      m_nRefVect[i] = a_nRefVect[i];
    }
  // will need to recompute local blockfactors in this case
  computeLocalBlockFactors();
}
  
/// sets fillRatio
void 
MeshRefine::fillRatio(const Real a_fill_ratio)
{
  m_fillRatio = a_fill_ratio;
}
  
/// sets blocking factor
void 
MeshRefine::blockFactor(const int a_block_factor)
{
  m_blockFactor = a_block_factor;
  // then need to recompute local blockFactors
  computeLocalBlockFactors();
}

/// sets proper nesting buffer size 
void 
MeshRefine::bufferSize(const int a_buffer_size)
{
  m_bufferSize = a_buffer_size;
}
  
/// sets maximum box size in any dimension
void  
MeshRefine::maxSize(const int a_max_size)
{
  m_maxSize = a_max_size;
}

/// has this object been defined properly?
bool 
MeshRefine::isDefined() const
{
  return m_isDefined;
}


