/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "HDF5.H"

using std::cout;
using std::endl;

/// Prototypes:

void
parseTestOptions( int argc ,char* argv[] ) ;

int
test();


/// Global variables for handling output:
static const char *pgmname = "HDF5attributes" ;
static const char *indent = "   ", *indent2 = "      " ; 
static bool verbose = true ;

int compare(const HDF5HeaderData& lhs, const HDF5HeaderData& rhs);

/// Code:

int 
main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif

  parseTestOptions( argc ,argv ) ;

  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  ///
  // Run the tests
  ///
  if(test() != 0)
    {
      cout << indent << pgmname <<" failed"<<endl;
    }
  else
    {
      cout << indent << pgmname <<" passed"<<endl;
    }
#ifdef MPI
  MPI_Finalize();
#endif
}

// returns 0 on all tests passed.

int test()
{
  int error;
  HDF5Handle testFile;

  assert(!testFile.isOpen());

  error = testFile.open("attributeTest.h5", HDF5Handle::CREATE);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "File creation failed "<<error<<endl;
      return error;
    }

  assert(testFile.isOpen()); 

  HDF5HeaderData set1;

  Real dx=0.004;
  Box b1(IntVect(D_DECL(1,2,1)), IntVect(D_DECL(4,4,4)));
  Box b2(IntVect(D_DECL(5,2,1)), IntVect(D_DECL(12,4,4)));
  int currentStep = 2332;

  set1.m_string["name"] = "set1";
  set1.m_real["dx"] = dx;
  set1.m_int["currentStep"] = currentStep;
  set1.m_intvect["some intvect or other"] = b1.smallEnd();
  set1.m_box["b1"] = b1;
  set1.m_box["b2"] = b2;

  error = set1.writeToFile(testFile); //write information to file root.
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "attribute write failed "<<error<<endl;
      return error;
    }
  testFile.setGroupToLevel(2);

  error = set1.writeToFile(testFile);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "attribute write failed "<<error<<endl;
      return error;
    }

  error = testFile.setGroup("/myspecialgroup");
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "set group failed "<<error<<endl;
      return error;
    }

  testFile.close();

  assert(!testFile.isOpen());
  //=================================================================

  HDF5HeaderData readData;

  error = testFile.open("attributeTest.h5", HDF5Handle::OPEN_RDONLY);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "File open failed "<<error<<endl;
      return error;
    }
  assert(testFile.isOpen());

  error = readData.readFromFile(testFile);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "attribute read failed "<<error<<endl;
      return error;
    }

  error = compare(set1, readData);
  if(error != 0)
    {
      if( verbose )
        {
          cout << indent2 << "attribute read does not match write in root"<<error<<"\n";
          cout << set1;
          cout <<"readData:\n";
          cout <<readData<<endl;
        }
      return error;
    }

  testFile.setGroupToLevel(2);

  readData.clear();

  if(readData.m_int.find("currentStep") != readData.m_int.end())
    {
      if( verbose )
        cout << indent2 << "HDF5HeaderData.clear() not functioning correct"<<endl;
      return 1;
    }

  testFile.close();

  return 0;

}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] ) 
{
  for( int i = 1 ; i < argc ; ++i )
  {
    if( argv[i][0] == '-' ) //if it is an option
    {
      // compare 3 chars to differentiate -x from -xx
      if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
      {
        verbose = true ;
        argv[i] = "" ;
      }
      else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
      {
        verbose = false ;
        argv[i] = "" ;
      }
      else
      {
        break ;
      }
    }
  }
  return ;
}

/** returns 0 if they are the same
            1 if an attribute exists in lhs that does not exist in rhs
            2 if an attribute exists in rhs that does not exist in lhs
            3 if a common attribute has a different value
*/
int compare(const HDF5HeaderData& lhs, const HDF5HeaderData& rhs)
{

#define LEFTVSRIGHT(left, right,  mapName, Ttype, code)                        \
  for(map<std::string, Ttype>::const_iterator p = left.mapName.begin();        \
      p!= left.mapName.end(); ++p){                                            \
      map<std::string, Ttype>::const_iterator  q = right.mapName.find(p->first);  \
      if(q == right.mapName.end())                                             \
      {                                                                        \
         return code;                                                          \
      }                                                                        \
      if(q->second != p->second) return 3; }

  LEFTVSRIGHT(lhs, rhs, m_real, Real, 1);
  LEFTVSRIGHT(lhs, rhs, m_int, int, 1);
  LEFTVSRIGHT(lhs, rhs, m_string, std::string, 1);
  LEFTVSRIGHT(lhs, rhs, m_intvect, IntVect, 1);
  LEFTVSRIGHT(lhs, rhs, m_box, Box, 1);

  LEFTVSRIGHT(rhs, lhs, m_real, Real, 2);
  LEFTVSRIGHT(rhs, lhs, m_int, int, 2);
  LEFTVSRIGHT(rhs, lhs, m_string, std::string, 2);
  LEFTVSRIGHT(rhs, lhs, m_intvect, IntVect, 2);
  LEFTVSRIGHT(rhs, lhs, m_box, Box, 2);

  return 0;
}


