/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <cstdio>
#include <cmath>
#include "Box.H"
#include "SPMD.H"
#include "REAL.H"
#include "MayDay.H"
#include "IntVectSet.H"
#include "BRMeshRefine.H"
#include "parstream.H"
#include <iostream>
#include <fstream>
#include <string>
#ifdef MPI
#include <mpi.h>
#endif

using std::endl;
using std::cout;
using std::cerr;

/// Global variables for test output:
static const char *pgmname = "domainSplitTest" ;
static const char *indent = "   " ,*indent2 = "      " ; 
static bool verbose = true ;

const int destProc = 0;

/// Prototypes:

void
parseTestOptions( int argc ,char* argv[] ) ;
/**
   domainSplitTest returns:
   0: all tests passed
   1: boxes coming out of domain split are not in 
   the original domain.
   2: boxes coming out of domain split have a side
   that is too long
   3: domainsplit returned an error code
   4: boxes do not cover  domain
  */
int domainSplitTest(void);



int main(int argc, char* argv[])
{
  using std::ostream;
  bool passed = true;
  int icode = 0;

#ifdef MPI
  MPI_Init(&argc, &argv);
#endif

  parseTestOptions( argc ,argv ) ;
  if( verbose )
    pout () << indent2 << "Beginning " << pgmname << " ..." << endl ;

  icode = domainSplitTest();

  if (procID() == destProc)
  {
      if(icode != 0)
      {
          cout << indent << pgmname 
               << ": failed with error code " << icode
               << endl;
          passed = false;
      }
      if (passed)
      {
          cout << indent << pgmname
               << ": passed all tests"
               << endl;
      }
  }

#ifdef MPI
  MPI_Finalize();
#endif

  return 0;
}
int domainSplitTest()
{
  int retflag = 0;
  int domlen = 100;
  int maxlen = 25;
  int blockf = 2;

  Box domain(IntVect::Zero, (domlen-1)*IntVect::Unit);

  Vector<Box> vbox;

  domainSplit(domain, vbox, maxlen, blockf);

  IntVectSet ivsDom(domain);
  for(int ibox = 0; ibox < vbox.size(); ibox++)
    {
      const Box& boxloc = vbox[ibox];
      if(!domain.contains(boxloc))
        {
          if(verbose)
            cerr << "error: " << boxloc << "not contained in " << domain  << endl;
          return 1;
        }
      ivsDom -= boxloc;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          if(boxloc.size(idir) > maxlen)
            {
              if(verbose)
                cerr << "error: " << boxloc << "is too big in " << idir 
                     << " direction (maxlen == " << maxlen << ")"  << endl;
              return 2;
            }
        }
    }
  if(!ivsDom.isEmpty())
    {
      if(verbose)
        cerr << "error: domain not covered " << endl;
      return 4;
    }
  if(verbose)
    {
      cout << "domain = " << domain << endl;
      cout << "max box size = " << maxlen << endl;
      cout << "boxes = " << endl;
      for(int ibox = 0; ibox < vbox.size(); ibox++)
        {
          const Box& boxloc = vbox[ibox];
          cout << boxloc << endl;
        }
      cout << "box sizes = " << endl;
      for(int ibox = 0; ibox < vbox.size(); ibox++)
        {
          const Box& boxloc = vbox[ibox];
          for(int idir = 0; idir < SpaceDim; idir ++)
            cout << boxloc.size(idir)  << "   ";
          cout << endl;
        }
    }

  return retflag;
} 

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] ) 
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}

