/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Tues, Oct 5 1999

#include <cmath>
#include "Misc.H"
#include "FArrayBox.H"
#include "BoxIterator.H"

#ifdef PGF90
// uncomment this line for pgf90...
//#include "fabTestF_F.H"
#endif

#ifdef MPI
#include <mpi.h>
#endif
/// Prototypes:
void
parseTestOptions( int argc ,char* argv[] ) ;

/**
   fabTest returns:
    0: all tests passed
   -1: setVal seems to be broken
   -2: operator*= seems to be broken
   -3: operator+= seems to be broken
   -4: operator-= seems to be broken
   -5: operator/= seems to be broken
   -6: L_\infty norm is broken
   -7: L_1 norm is broken
   -8: L_2 norm is broken
   -9: linearIn, linearOut, size() is broken
 */
extern int
fabTest(void);


/// Global variables for handling output:
static const char *pgmname = "fabTest" ;
static const char *indent = "   ", *indent2 = "      " ; 
static bool verbose = true ;

#ifdef PGF90
// define main loop here.
static int zz = 0;
extern "C" void pghpf_init(int *);
int __argc_save;
char **__argv_save;
#endif

/// Code:
int
main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif

#ifdef PGF90
  __argc_save = argc;  
  __argv_save = argv; 

  pghpf_init(&zz);  
#endif

  bool passed = true;
  int icode = 0;

  parseTestOptions( argc ,argv ) ;
  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

#ifdef PGF90
  FORT_F90FABTEST(CHF_INT(icode));
#endif

  icode = fabTest();
  if(icode != 0)
    {
      cout << indent << pgmname
           << ": FArrayBox test failed with error code " << icode
           << endl;
      passed = false;
    }

  if(passed)
      cout << indent << pgmname
           << ": FArrayBox passed all tests"
           << endl;

#ifdef MPI
  MPI_Finalize();
#endif
  return 0;
}

int fabTest()
{
  //this is the error code that will get returned if a test fails
  int icode = 0;

  int ihigh = 8;
  Real eps = 1.0e-9;
  //test 1.  make a box of a certain size.  set it to one.  see if 
  //add, subtract, multiply divide
  IntVect hiVect = (ihigh-1)*IntVect::TheUnitVector();
  IntVect loVect(D_DECL(2,3,4));
  Box B(loVect, hiVect);
  FArrayBox fabOne(B,1);
  FArrayBox fabTwo(B,1);
  ///setVal test
  fabOne.setVal(1.);

  if(((fabOne.max() > 1.0+eps)||
      (fabOne.max() < 1.0-eps)) ||
     ((fabOne.min() > 1.0+eps)||
      (fabOne.min() < 1.0-eps)))
    {
      cout << indent << pgmname
           << ": fab::setval might have failed"
           << endl ;
      icode = -1;
    }
  else if(verbose)
    {
      cout << indent2 << pgmname
           << ": fab::setval passed"
           << endl ;
    }

  ///arithmetic operator tests
  fabTwo.copy(fabOne);
  fabTwo *= 2.0;
  if(((fabTwo.max() > 2.0+eps)||
      (fabTwo.max() < 2.0-eps)) ||
     ((fabTwo.min() > 2.0+eps)||
      (fabTwo.min() < 2.0-eps)))
    {
      cout << indent << pgmname
           << ": fab::operator*= might have failed"
           << endl ;
      icode = -2;
      return(icode);
    }
  else if(verbose)
    {
      cout << indent2 << pgmname
           << ": fab::operator*= passed"
           << endl ;
    }


  FArrayBox fabThree(B, 1);
  fabThree.copy(fabTwo);
  fabThree += 1.0;
  if(((fabThree.max() > 3.0+eps)||
      (fabThree.max() < 3.0-eps)) ||
     ((fabThree.min() > 3.0+eps)||
      (fabThree.min() < 3.0-eps)))
    {
      cout << indent << pgmname
           << ": fab::operator+= might have failed"
           << endl ;
      icode = -3;
      return(icode);
    }
  else if(verbose)
    {
      cout << indent2 << pgmname
           << ": fab::operator+= passed"
           << endl ;
    }


  fabTwo.copy(fabThree);
  fabTwo -= fabOne;
  if(((fabTwo.max() > 2.0+eps)||
      (fabTwo.max() < 2.0-eps)) ||
     ((fabTwo.min() > 2.0+eps)||
      (fabTwo.min() < 2.0-eps)))
    {
      cout << indent << pgmname
           << ": fab::operator-= might have failed"
           << endl ;
      icode = -4;
      return(icode);
    }
  else if(verbose)
    {
      cout << indent2 << pgmname
           << ": fab::operator-= passed"
           << endl ;
    }


  FArrayBox fabFour(B,1);
  fabFour.setVal(8.0);
  fabFour /= fabTwo;
  if(((fabFour.max() > 4.0+eps)||
      (fabFour.max() < 4.0-eps)) ||
     ((fabFour.min() > 4.0+eps)||
      (fabFour.min() < 4.0-eps)))
    {
      cout << indent << pgmname
           << ": fab::operator/= might have failed"
           << endl ;
      icode = -5;
      return(icode);
    }
  else if(verbose)
    {
      cout << indent2 << pgmname
           << ": fab::operator/= passed"
           << endl ;
    }

  ///norm tests
  if((fabOne.norm(0,0,1) - 1.0 > eps) ||
     (fabTwo.norm(0,0,1) - 2.0 > eps) ||
     (fabThree.norm(0,0,1) - 3.0 > eps) ||
     (fabFour.norm(0,0,1) - 4.0 > eps))
    {
      cout << indent << pgmname
           << ": fab::norm(0) might have failed"
           << endl ;
      icode = -6;
      return(icode);
    }
  else if(verbose)
    {
      cout << indent2 << pgmname
           << ": fab::norm(0) passed"
           << endl ;
    }

  Real rnorm = B.numPts();
  if((fabOne.norm(1,0,1)/rnorm - 1.0 > eps) ||
     (fabTwo.norm(1,0,1)/rnorm - 2.0 > eps) ||
     (fabThree.norm(1,0,1)/rnorm - 3.0 > eps) ||
     (fabFour.norm(1,0,1)/rnorm - 4.0 > eps))
    {
      cout << indent << pgmname
           << ": fab::norm(1) might have failed"
           << endl ;
      icode = -7;
      return(icode);
    }
  else if(verbose)
    {
      cout << indent2 << pgmname
           << ": fab::norm(1) passed"
           << endl ;
    }

  rnorm = sqrt(rnorm);
  if((fabOne.norm(2,0,1)/rnorm - 1.0 > eps) ||
     (fabTwo.norm(2,0,1)/rnorm - 2.0 > eps) ||
     (fabThree.norm(2,0,1)/rnorm - 3.0 > eps) ||
     (fabFour.norm(2,0,1)/rnorm - 4.0 > eps))
    {
      cout << indent << pgmname
           << ": fab::norm(2) might have failed"
           << endl ;
      icode = -8;
      return(icode);
    }
  else if(verbose)
    {
      cout << indent2 << pgmname
           << ": fab::norm(2) passed"
           << endl ;
    }

  // ======linearization tests=====================================
  Box B2(loVect+2*IntVect::TheUnitVector(), hiVect);
  assert(!B2.isEmpty());

  BaseFab<Real> intfab1(B, 2), intfab2(B, 3), intfab3(B2, 1);

  intfab1.setVal(-11);
  intfab2.setVal(2);
  intfab3.setVal(3);

  Interval interval1(0, 1);
  Interval interval2(0, 0);
  Interval interval3(2, 2);
  int sz = intfab1.size(B2, interval1);
  void* buf = malloc(sz);
  assert(buf != NULL);
  intfab1.linearOut(buf, B2, interval1);
  intfab2.linearIn(buf, B2, interval1);

  for(BoxIterator bit(B); bit.ok(); ++bit)
    { // check that non-modified data actually is non-modifed
      if(!(intfab1(bit(), 0) == -11 && 
           intfab1(bit(), 1) == -11 &&
           intfab2(bit(), 2) == 2))
        {
          cout << indent << pgmname
               << ": BaseFab linearization test #1 might have failed."
               << endl ;
          return -9;
        }
    }
  for(BoxIterator bit(B2); bit.ok(); ++bit)
    { //check that modified data is properly modified.
      if(!(intfab2(bit(), 0) == -11 &&
           intfab2(bit(), 1) == -11))
        {
          cout << indent << pgmname
               << ": BaseFab linearization test #2 might have failed"
               << endl ;
          return -9;
        }
    }

  int sz2 = intfab3.size(B2, interval2);
  //reuse buffer here if possible
  if(sz2 > sz)
    {
      free(buf);
      buf = malloc(sz2);
    }
  assert(buf != NULL);
  intfab3.linearOut(buf, B2, interval2);
  intfab2.linearIn(buf, B2, interval3);

  for(BoxIterator bit(B2); bit.ok(); ++bit)
    {
      if(!(intfab3(bit(), 0) == 3 &&
           intfab2(bit(), 2) == 3))
        {
          cout << indent << pgmname
               << ": BaseFab linearization test #3 might have failed"
               << endl ;
          return -9;
        }
    }

  free(buf);

  //=======end linearization tests =========================================


  return(icode);
} 


///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] ) 
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}
